
template <typename Tuple, std::size_t ... Is>
auto pop_front2_impl(const Tuple& tuple, std::index_sequence<Is...>)
{
    return std::make_tuple(std::get<2 + Is>(tuple)...);
}

template <typename Tuple>
auto pop_front2(const Tuple& tuple)
{
    return pop_front2_impl(tuple,
                          std::make_index_sequence<std::tuple_size<Tuple>::value - 2>());
}

template<int N1, int N2>
auto Concat2(SIMD<double,N1> simd1, SIMD<double,N2> simd2)
{
    if constexpr (IsNativeSIMDSize(simd1.Size())) {
        return SIMD<double,simd1.Size()+simd2.Size()>(simd1, simd2);
    }
    else {
      auto lo = simd1.Lo();
      auto hi = simd1.Hi();
  
      SIMD<double,hi.Size()+simd2.Size()> res1(hi, simd2);
      SIMD<double,simd1.Size()+simd2.Size()> res2(lo, res1);
      return res2;
    }
}

template <typename ...Args, int N>
auto Concat (tuple<SIMD<double,N>, Args...> tup)
{
  if constexpr (tuple_size<tuple<SIMD<double,N>, Args...>>() == 1)
                 return get<0>(tup);
  else if constexpr (tuple_size<tuple<SIMD<double,N>, Args...>>() == 2)
      return Concat2(get<0>(tup), get<1>(tup));
  else
    {
      auto front = Concat2(get<0>(tup), get<1>(tup));
      auto rest = pop_front2(tup);
      return Concat(std::tuple_cat(make_tuple(front), rest));
    }
}

template <int N>
void FMAnonasm (SIMD<double,N> a, SIMD<double,N> b, SIMD<double,N> & sum)
{ sum = FMA(a,b,sum); } template <int N>
void FNMAnonasm (SIMD<double,N> a, SIMD<double,N> b, SIMD<double,N> & sum)
{ sum = FNMA(a,b,sum); }static_assert(SIMD<double>::Size() == 1, "inconsistent compile flags for generate_mat_kernels.cpp and matkernel.hpp");
enum OPERATION { ADD, SUB, SET, SETNEG };
 /* *********************** MatKernelMultAB ********************* */
 /* A,B,C ... row major storage                                   */
 /* dim C = H * (SW*W)     SW .. SIMD<double>::Size()             */
 /* OP == SET:    C = A * B                                       */
 /* OP == ADD:    C += A * B                                      */
 /* OP == NEG:    C = -A * B                                      */
 /* OP == SUB:    C -= A * B                                      */
 /* ************************************************************* */
template <size_t H, size_t W, OPERATION OP>
inline void MatKernelMultAB
(size_t n, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <size_t H, size_t W, OPERATION OP>
inline void MatKernelMultAB
(size_t n, double * pa, size_t da, SIMD<double> * pb, size_t db, double * pc, size_t dc);
template <size_t H, size_t W>
inline void MatKernelAlignedMultAB
(size_t n, double * pa, size_t da, SIMD<double> * pb, size_t db, SIMD<double> * pc, size_t dc);
template <> INLINE void MatKernelMultAB<1, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
}
sum00.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<1, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
}
pc[0]= sum00;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<2, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
}
pc[0]= sum00;
pc += dc;
pc[0]= sum10;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<3, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
}
pc[0]= sum00;
pc += dc;
pc[0]= sum10;
pc += dc;
pc[0]= sum20;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<4, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
}
pc[0]= sum00;
pc += dc;
pc[0]= sum10;
pc += dc;
pc[0]= sum20;
pc += dc;
pc[0]= sum30;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<5, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
}
pc[0]= sum00;
pc += dc;
pc[0]= sum10;
pc += dc;
pc[0]= sum20;
pc += dc;
pc[0]= sum30;
pc += dc;
pc[0]= sum40;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<6, 1>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
}
pc[0]= sum00;
pc += dc;
pc[0]= sum10;
pc += dc;
pc[0]= sum20;
pc += dc;
pc[0]= sum30;
pc += dc;
pc[0]= sum40;
pc += dc;
pc[0]= sum50;
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<1, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<2, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<3, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<4, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<5, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<6, 2>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
}
pc[0]= sum00;
pc[1]= sum01;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc += dc;
pc[0]= sum50;
pc[1]= sum51;
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<1, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<2, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<3, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<4, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<5, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc[2]= sum42;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<6, 3>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc[2]= sum42;
pc += dc;
pc[0]= sum50;
pc[1]= sum51;
pc[2]= sum52;
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<1, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<2, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<3, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<4, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<5, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
FMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
FNMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
SIMD<double> sum53(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
FMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
SIMD<double> sum53(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> b1(pb+1*SW);
SIMD<double> b2(pb+2*SW);
SIMD<double> b3(pb+3*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
FNMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
FMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
FNMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
SIMD<double> sum53(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
FMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> INLINE void MatKernelMultAB<6, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
SIMD<double> sum53(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> b1 = pb[1];
SIMD<double> b2 = pb[2];
SIMD<double> b3 = pb[3];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
FNMAasm(a0,b1,sum01);
FNMAasm(a0,b2,sum02);
FNMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
FNMAasm(a1,b1,sum11);
FNMAasm(a1,b2,sum12);
FNMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
FNMAasm(a2,b1,sum21);
FNMAasm(a2,b2,sum22);
FNMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
FNMAasm(a3,b1,sum31);
FNMAasm(a3,b2,sum32);
FNMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
FNMAasm(a4,b1,sum41);
FNMAasm(a4,b2,sum42);
FNMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
FNMAasm(a5,b1,sum51);
FNMAasm(a5,b2,sum52);
FNMAasm(a5,b3,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<1, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<2, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc[3]= sum13;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<3, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc[3]= sum13;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc[3]= sum23;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<4, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc[3]= sum13;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc[3]= sum23;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc[3]= sum33;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<5, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc[3]= sum13;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc[3]= sum23;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc[3]= sum33;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc[2]= sum42;
pc[3]= sum43;
pc += dc;
}
template <> inline void MatKernelAlignedMultAB<6, 4>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     SIMD<double> * pc, size_t dc)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb[0]);
SIMD<double> b1(pb[1]);
SIMD<double> b2(pb[2]);
SIMD<double> b3(pb[3]);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
FMAasm(a0,b1,sum01);
FMAasm(a0,b2,sum02);
FMAasm(a0,b3,sum03);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
FMAasm(a1,b1,sum11);
FMAasm(a1,b2,sum12);
FMAasm(a1,b3,sum13);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
FMAasm(a2,b1,sum21);
FMAasm(a2,b2,sum22);
FMAasm(a2,b3,sum23);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
FMAasm(a3,b1,sum31);
FMAasm(a3,b2,sum32);
FMAasm(a3,b3,sum33);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
FMAasm(a4,b1,sum41);
FMAasm(a4,b2,sum42);
FMAasm(a4,b3,sum43);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
FMAasm(a5,b1,sum51);
FMAasm(a5,b2,sum52);
FMAasm(a5,b3,sum53);
}
pc[0]= sum00;
pc[1]= sum01;
pc[2]= sum02;
pc[3]= sum03;
pc += dc;
pc[0]= sum10;
pc[1]= sum11;
pc[2]= sum12;
pc[3]= sum13;
pc += dc;
pc[0]= sum20;
pc[1]= sum21;
pc[2]= sum22;
pc[3]= sum23;
pc += dc;
pc[0]= sum30;
pc[1]= sum31;
pc[2]= sum32;
pc[3]= sum33;
pc += dc;
pc[0]= sum40;
pc[1]= sum41;
pc[2]= sum42;
pc[3]= sum43;
pc += dc;
pc[0]= sum50;
pc[1]= sum51;
pc[2]= sum52;
pc[3]= sum53;
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<8, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
SIMD<double> sum80(0);
SIMD<double> sum90(0);
SIMD<double> sum100(0);
SIMD<double> sum110(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
SIMD<double> sum80(0);
SIMD<double> sum90(0);
SIMD<double> sum100(0);
SIMD<double> sum110(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FNMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FNMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FNMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FNMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
SIMD<double> sum80(pc+SW*0);
pc += dc;
SIMD<double> sum90(pc+SW*0);
pc += dc;
SIMD<double> sum100(pc+SW*0);
pc += dc;
SIMD<double> sum110(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
SIMD<double> sum80(pc+SW*0);
pc += dc;
SIMD<double> sum90(pc+SW*0);
pc += dc;
SIMD<double> sum100(pc+SW*0);
pc += dc;
SIMD<double> sum110(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0(pb+0*SW);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FNMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FNMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FNMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FNMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SET>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
SIMD<double> sum80(0);
SIMD<double> sum90(0);
SIMD<double> sum100(0);
SIMD<double> sum110(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SETNEG>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
SIMD<double> sum80(0);
SIMD<double> sum90(0);
SIMD<double> sum100(0);
SIMD<double> sum110(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FNMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FNMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FNMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FNMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
SIMD<double> sum80(pc+SW*0);
pc += dc;
SIMD<double> sum90(pc+SW*0);
pc += dc;
SIMD<double> sum100(pc+SW*0);
pc += dc;
SIMD<double> sum110(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <> INLINE void MatKernelMultAB<12, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
pc += dc;
SIMD<double> sum10(pc+SW*0);
pc += dc;
SIMD<double> sum20(pc+SW*0);
pc += dc;
SIMD<double> sum30(pc+SW*0);
pc += dc;
SIMD<double> sum40(pc+SW*0);
pc += dc;
SIMD<double> sum50(pc+SW*0);
pc += dc;
SIMD<double> sum60(pc+SW*0);
pc += dc;
SIMD<double> sum70(pc+SW*0);
pc += dc;
SIMD<double> sum80(pc+SW*0);
pc += dc;
SIMD<double> sum90(pc+SW*0);
pc += dc;
SIMD<double> sum100(pc+SW*0);
pc += dc;
SIMD<double> sum110(pc+SW*0);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b0 = pb[0];
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b0,sum00);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b0,sum10);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b0,sum20);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b0,sum30);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b0,sum40);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b0,sum50);
SIMD<double> a6(pa[6*da]);
FNMAasm(a6,b0,sum60);
SIMD<double> a7(pa[7*da]);
FNMAasm(a7,b0,sum70);
SIMD<double> a8(pa[8*da]);
FNMAasm(a8,b0,sum80);
SIMD<double> a9(pa[9*da]);
FNMAasm(a9,b0,sum90);
SIMD<double> a10(pa[10*da]);
FNMAasm(a10,b0,sum100);
SIMD<double> a11(pa[11*da]);
FNMAasm(a11,b0,sum110);
}
sum00.Store(pc+SW*0);
pc += dc;
sum10.Store(pc+SW*0);
pc += dc;
sum20.Store(pc+SW*0);
pc += dc;
sum30.Store(pc+SW*0);
pc += dc;
sum40.Store(pc+SW*0);
pc += dc;
sum50.Store(pc+SW*0);
pc += dc;
sum60.Store(pc+SW*0);
pc += dc;
sum70.Store(pc+SW*0);
pc += dc;
sum80.Store(pc+SW*0);
pc += dc;
sum90.Store(pc+SW*0);
pc += dc;
sum100.Store(pc+SW*0);
pc += dc;
sum110.Store(pc+SW*0);
pc += dc;
}
template <size_t H, OPERATION OP>
inline void MatKernelMultABMask
(size_t n, SIMD<mask64> mask, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <size_t H, OPERATION OP>
inline void MatKernelMultABMask
(size_t n, SIMD<mask64> mask, double * pa, size_t da, SIMD<double> * pb, size_t db, double * pc, size_t dc);
template <> inline void MatKernelMultABMask<1, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<1, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
}
sum0.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<2, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<3, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<4, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<5, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
SIMD<double> sum5(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
SIMD<double> sum5(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
SIMD<double> sum5(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
SIMD<double> sum5(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SET>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
SIMD<double> sum5(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SETNEG>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
SIMD<double> sum0(0);
SIMD<double> sum1(0);
SIMD<double> sum2(0);
SIMD<double> sum3(0);
SIMD<double> sum4(0);
SIMD<double> sum5(0);
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, ADD>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
SIMD<double> sum5(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
template <> inline void MatKernelMultABMask<6, SUB>
    (size_t n, SIMD<mask64> mask,
     double * pa, size_t da,
     SIMD<double> * pb, size_t db,
     double * pc, size_t dc)
{
double * hpc = pc;
SIMD<double> sum0(pc, mask);
pc += dc;
SIMD<double> sum1(pc, mask);
pc += dc;
SIMD<double> sum2(pc, mask);
pc += dc;
SIMD<double> sum3(pc, mask);
pc += dc;
SIMD<double> sum4(pc, mask);
pc += dc;
SIMD<double> sum5(pc, mask);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa++, pb += db) {
SIMD<double> b((double*)pb,mask);
SIMD<double> a0(pa[0*da]);
FNMAasm(a0,b,sum0);
SIMD<double> a1(pa[1*da]);
FNMAasm(a1,b,sum1);
SIMD<double> a2(pa[2*da]);
FNMAasm(a2,b,sum2);
SIMD<double> a3(pa[3*da]);
FNMAasm(a3,b,sum3);
SIMD<double> a4(pa[4*da]);
FNMAasm(a4,b,sum4);
SIMD<double> a5(pa[5*da]);
FNMAasm(a5,b,sum5);
}
sum0.Store(pc,mask);
pc += dc;
sum1.Store(pc,mask);
pc += dc;
sum2.Store(pc,mask);
pc += dc;
sum3.Store(pc,mask);
pc += dc;
sum4.Store(pc,mask);
pc += dc;
sum5.Store(pc,mask);
pc += dc;
}
 /* *********************** MatKernelScalAB ********************* */
 /* Inner products of rows of A with rows of B                    */
 /* A,B ... row major storage                                     */
 /* dim A = H * n                                                 */
 /* dim B = W * n                                                 */
 /* ************************************************************* */
template <size_t H, size_t W> inline auto MatKernelScalAB
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db);
template <size_t H, size_t W> inline auto MatKernelScalAB
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db);
template <> INLINE auto MatKernelScalAB<6, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> a3(pa+3*da+i);
SIMD<double> a4(pa+4*da+i);
SIMD<double> a5(pa+5*da+i);
SIMD<double> b0(pb+0*db+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb+1*db+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
SIMD<double> b2(pb+2*db+i);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
FMAasm(a3,b2,sum32);
FMAasm(a4,b2,sum42);
FMAasm(a5,b2,sum52);
SIMD<double> b3(pb+3*db+i);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
FMAasm(a3,b3,sum33);
FMAasm(a4,b3,sum43);
FMAasm(a5,b3,sum53);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> a3(pa+3*da+i, mask);
SIMD<double> a4(pa+4*da+i, mask);
SIMD<double> a5(pa+5*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb+1*db+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
SIMD<double> b2(pb+2*db+i, mask);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
FMAasm(a3,b2,sum32);
FMAasm(a4,b2,sum42);
FMAasm(a5,b2,sum52);
SIMD<double> b3(pb+3*db+i, mask);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
FMAasm(a3,b3,sum33);
FMAasm(a4,b3,sum43);
FMAasm(a5,b3,sum53);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23),HSum(sum30,sum31,sum32,sum33),HSum(sum40,sum41,sum42,sum43),HSum(sum50,sum51,sum52,sum53));
}
template <> INLINE auto MatKernelScalAB<6, 4>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> a3(pa[3*da+i]);
SIMD<double> a4(pa[4*da+i]);
SIMD<double> a5(pa[5*da+i]);
SIMD<double> b0(pb[0*db+i]);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb[1*db+i]);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
SIMD<double> b2(pb[2*db+i]);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
FMAasm(a3,b2,sum32);
FMAasm(a4,b2,sum42);
FMAasm(a5,b2,sum52);
SIMD<double> b3(pb[3*db+i]);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
FMAasm(a3,b3,sum33);
FMAasm(a4,b3,sum43);
FMAasm(a5,b3,sum53);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23),HSum(sum30,sum31,sum32,sum33),HSum(sum40,sum41,sum42,sum43),HSum(sum50,sum51,sum52,sum53));
}
template <> INLINE auto MatKernelScalAB<3, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> b0(pb+0*db+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb+1*db+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
SIMD<double> b2(pb+2*db+i);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
SIMD<double> b3(pb+3*db+i);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb+1*db+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
SIMD<double> b2(pb+2*db+i, mask);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
SIMD<double> b3(pb+3*db+i, mask);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23));
}
template <> INLINE auto MatKernelScalAB<3, 4>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> b0(pb[0*db+i]);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb[1*db+i]);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
SIMD<double> b2(pb[2*db+i]);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
SIMD<double> b3(pb[3*db+i]);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23));
}
template <> INLINE auto MatKernelScalAB<1, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
SIMD<double> b1(pb+1*db+i);
sum01 += a0 * b1;
SIMD<double> b2(pb+2*db+i);
sum02 += a0 * b2;
SIMD<double> b3(pb+3*db+i);
sum03 += a0 * b3;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
SIMD<double> b1(pb+1*db+i, mask);
FMAasm(a0,b1,sum01);
SIMD<double> b2(pb+2*db+i, mask);
FMAasm(a0,b2,sum02);
SIMD<double> b3(pb+3*db+i, mask);
FMAasm(a0,b3,sum03);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03));
}
template <> INLINE auto MatKernelScalAB<1, 4>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
SIMD<double> b1(pb[1*db+i]);
sum01 += a0 * b1;
SIMD<double> b2(pb[2*db+i]);
sum02 += a0 * b2;
SIMD<double> b3(pb[3*db+i]);
sum03 += a0 * b3;
}
return make_tuple(HSum(sum00,sum01,sum02,sum03));
}
template <> INLINE auto MatKernelScalAB<6, 2>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> a3(pa+3*da+i);
SIMD<double> a4(pa+4*da+i);
SIMD<double> a5(pa+5*da+i);
SIMD<double> b0(pb+0*db+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb+1*db+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> a3(pa+3*da+i, mask);
SIMD<double> a4(pa+4*da+i, mask);
SIMD<double> a5(pa+5*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb+1*db+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21),HSum(sum30,sum31),HSum(sum40,sum41),HSum(sum50,sum51));
}
template <> INLINE auto MatKernelScalAB<6, 2>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> a3(pa[3*da+i]);
SIMD<double> a4(pa[4*da+i]);
SIMD<double> a5(pa[5*da+i]);
SIMD<double> b0(pb[0*db+i]);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb[1*db+i]);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21),HSum(sum30,sum31),HSum(sum40,sum41),HSum(sum50,sum51));
}
template <> INLINE auto MatKernelScalAB<3, 2>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
SIMD<double> b1(pb+1*db+i);
sum01 += a0 * b1;
sum11 += a1 * b1;
sum21 += a2 * b1;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb+1*db+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21));
}
template <> INLINE auto MatKernelScalAB<3, 2>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
SIMD<double> b1(pb[1*db+i]);
sum01 += a0 * b1;
sum11 += a1 * b1;
sum21 += a2 * b1;
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21));
}
template <> INLINE auto MatKernelScalAB<8, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> a3(pa+3*da+i);
SIMD<double> a4(pa+4*da+i);
SIMD<double> a5(pa+5*da+i);
SIMD<double> a6(pa+6*da+i);
SIMD<double> a7(pa+7*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
sum60 += a6 * b0;
sum70 += a7 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> a3(pa+3*da+i, mask);
SIMD<double> a4(pa+4*da+i, mask);
SIMD<double> a5(pa+5*da+i, mask);
SIMD<double> a6(pa+6*da+i, mask);
SIMD<double> a7(pa+7*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
FMAasm(a6,b0,sum60);
FMAasm(a7,b0,sum70);
}
return Concat(make_tuple(HSum(sum00, sum10, sum20, sum30),HSum(sum40, sum50, sum60, sum70)));
}
template <> INLINE auto MatKernelScalAB<8, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> a3(pa[3*da+i]);
SIMD<double> a4(pa[4*da+i]);
SIMD<double> a5(pa[5*da+i]);
SIMD<double> a6(pa[6*da+i]);
SIMD<double> a7(pa[7*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
sum60 += a6 * b0;
sum70 += a7 * b0;
}
return Concat(make_tuple(HSum(sum00, sum10, sum20, sum30),HSum(sum40, sum50, sum60, sum70)));
}
template <> INLINE auto MatKernelScalAB<6, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> a3(pa+3*da+i);
SIMD<double> a4(pa+4*da+i);
SIMD<double> a5(pa+5*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> a3(pa+3*da+i, mask);
SIMD<double> a4(pa+4*da+i, mask);
SIMD<double> a5(pa+5*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20),HSum(sum30),HSum(sum40),HSum(sum50));
}
template <> INLINE auto MatKernelScalAB<6, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> a3(pa[3*da+i]);
SIMD<double> a4(pa[4*da+i]);
SIMD<double> a5(pa[5*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20),HSum(sum30),HSum(sum40),HSum(sum50));
}
template <> INLINE auto MatKernelScalAB<4, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> a3(pa+3*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> a3(pa+3*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
}
return Concat(make_tuple(HSum(sum00, sum10, sum20, sum30)));
}
template <> INLINE auto MatKernelScalAB<4, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> a3(pa[3*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
}
return Concat(make_tuple(HSum(sum00, sum10, sum20, sum30)));
}
template <> INLINE auto MatKernelScalAB<3, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> a2(pa+2*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> a2(pa+2*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20));
}
template <> INLINE auto MatKernelScalAB<3, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> a2(pa[2*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20));
}
template <> INLINE auto MatKernelScalAB<2, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> a1(pa+1*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> a1(pa+1*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
}
return make_tuple(HSum(sum00),HSum(sum10));
}
template <> INLINE auto MatKernelScalAB<2, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
SIMD<double> sum10(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> a1(pa[1*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
sum10 += a1 * b0;
}
return make_tuple(HSum(sum00),HSum(sum10));
}
template <> INLINE auto MatKernelScalAB<1, 1>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa+0*da+i);
SIMD<double> b0(pb+0*db+i);
sum00 += a0 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa+0*da+i, mask);
SIMD<double> b0(pb+0*db+i, mask);
FMAasm(a0,b0,sum00);
}
return make_tuple(HSum(sum00));
}
template <> INLINE auto MatKernelScalAB<1, 1>
    (size_t n,
     SIMD<double> * pa, size_t da,
     SIMD<double> * pb, size_t db)
{
SIMD<double> sum00(0);
size_t i = 0;
for ( ; i < n; i++) {
SIMD<double> a0(pa[0*da+i]);
SIMD<double> b0(pb[0*db+i]);
sum00 += a0 * b0;
}
return make_tuple(HSum(sum00));
}
template <size_t H, size_t W> inline auto MultiVecScalAB
    (size_t n,
     double ** ppa,
     double ** ppb);
template <> INLINE auto MultiVecScalAB<6, 4>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum32(0);
SIMD<double> sum33(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum42(0);
SIMD<double> sum43(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
SIMD<double> sum52(0);
SIMD<double> sum53(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb1+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
SIMD<double> b2(pb2+i);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
FMAasm(a3,b2,sum32);
FMAasm(a4,b2,sum42);
FMAasm(a5,b2,sum52);
SIMD<double> b3(pb3+i);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
FMAasm(a3,b3,sum33);
FMAasm(a4,b3,sum43);
FMAasm(a5,b3,sum53);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb1+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
SIMD<double> b2(pb2+i, mask);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
FMAasm(a3,b2,sum32);
FMAasm(a4,b2,sum42);
FMAasm(a5,b2,sum52);
SIMD<double> b3(pb3+i, mask);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
FMAasm(a3,b3,sum33);
FMAasm(a4,b3,sum43);
FMAasm(a5,b3,sum53);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23),HSum(sum30,sum31,sum32,sum33),HSum(sum40,sum41,sum42,sum43),HSum(sum50,sum51,sum52,sum53));
}
template <> INLINE auto MultiVecScalAB<3, 4>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum12(0);
SIMD<double> sum13(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum22(0);
SIMD<double> sum23(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> b0(pb0+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb1+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
SIMD<double> b2(pb2+i);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
SIMD<double> b3(pb3+i);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb1+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
SIMD<double> b2(pb2+i, mask);
FMAasm(a0,b2,sum02);
FMAasm(a1,b2,sum12);
FMAasm(a2,b2,sum22);
SIMD<double> b3(pb3+i, mask);
FMAasm(a0,b3,sum03);
FMAasm(a1,b3,sum13);
FMAasm(a2,b3,sum23);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03),HSum(sum10,sum11,sum12,sum13),HSum(sum20,sum21,sum22,sum23));
}
template <> INLINE auto MultiVecScalAB<1, 4>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum02(0);
SIMD<double> sum03(0);
double* pa0 = ppa[0];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
SIMD<double> b1(pb1+i);
sum01 += a0 * b1;
SIMD<double> b2(pb2+i);
sum02 += a0 * b2;
SIMD<double> b3(pb3+i);
sum03 += a0 * b3;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
SIMD<double> b1(pb1+i, mask);
FMAasm(a0,b1,sum01);
SIMD<double> b2(pb2+i, mask);
FMAasm(a0,b2,sum02);
SIMD<double> b3(pb3+i, mask);
FMAasm(a0,b3,sum03);
}
return make_tuple(HSum(sum00,sum01,sum02,sum03));
}
template <> INLINE auto MultiVecScalAB<6, 2>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
SIMD<double> sum30(0);
SIMD<double> sum31(0);
SIMD<double> sum40(0);
SIMD<double> sum41(0);
SIMD<double> sum50(0);
SIMD<double> sum51(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb1+i);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
SIMD<double> b1(pb1+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
FMAasm(a3,b1,sum31);
FMAasm(a4,b1,sum41);
FMAasm(a5,b1,sum51);
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21),HSum(sum30,sum31),HSum(sum40,sum41),HSum(sum50,sum51));
}
template <> INLINE auto MultiVecScalAB<3, 2>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum01(0);
SIMD<double> sum10(0);
SIMD<double> sum11(0);
SIMD<double> sum20(0);
SIMD<double> sum21(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
SIMD<double> b1(pb1+i);
sum01 += a0 * b1;
sum11 += a1 * b1;
sum21 += a2 * b1;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
SIMD<double> b1(pb1+i, mask);
FMAasm(a0,b1,sum01);
FMAasm(a1,b1,sum11);
FMAasm(a2,b1,sum21);
}
return make_tuple(HSum(sum00,sum01),HSum(sum10,sum11),HSum(sum20,sum21));
}
template <> INLINE auto MultiVecScalAB<8, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
SIMD<double> sum60(0);
SIMD<double> sum70(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pa6 = ppa[6];
double* pa7 = ppa[7];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> a6(pa6+i);
SIMD<double> a7(pa7+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
sum60 += a6 * b0;
sum70 += a7 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> a6(pa6+i, mask);
SIMD<double> a7(pa7+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
FMAasm(a6,b0,sum60);
FMAasm(a7,b0,sum70);
}
return make_tuple(HSum(sum00, sum10, sum20, sum30),HSum(sum40, sum50, sum60, sum70));
}
template <> INLINE auto MultiVecScalAB<6, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
SIMD<double> sum40(0);
SIMD<double> sum50(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
sum40 += a4 * b0;
sum50 += a5 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
FMAasm(a4,b0,sum40);
FMAasm(a5,b0,sum50);
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20),HSum(sum30),HSum(sum40),HSum(sum50));
}
template <> INLINE auto MultiVecScalAB<4, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
SIMD<double> sum30(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
sum30 += a3 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
FMAasm(a3,b0,sum30);
}
return make_tuple(HSum(sum00, sum10, sum20, sum30));
}
template <> INLINE auto MultiVecScalAB<3, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
SIMD<double> sum20(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
sum20 += a2 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
FMAasm(a2,b0,sum20);
}
return make_tuple(HSum(sum00),HSum(sum10),HSum(sum20));
}
template <> INLINE auto MultiVecScalAB<2, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
SIMD<double> sum10(0);
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
sum10 += a1 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
FMAasm(a1,b0,sum10);
}
return make_tuple(HSum(sum00),HSum(sum10));
}
template <> INLINE auto MultiVecScalAB<1, 1>
    (size_t n,
     double** ppa,
     double** ppb)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> sum00(0);
double* pa0 = ppa[0];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
sum00 += a0 * b0;
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
FMAasm(a0,b0,sum00);
}
return make_tuple(HSum(sum00));
}
template <size_t H, size_t W, bool conjugate> inline void MultiVecScalC
    (size_t n,
     Complex ** ppa,
     Complex ** ppb,
     Complex* pc, size_t dc);
template <> INLINE void MultiVecScalC<6, 2, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum1_0(0);
SIMD<double> sum1_1(0);
SIMD<double> sum2_0(0);
SIMD<double> sum2_1(0);
SIMD<double> sum3_0(0);
SIMD<double> sum3_1(0);
SIMD<double> sum4_0(0);
SIMD<double> sum4_1(0);
SIMD<double> sum5_0(0);
SIMD<double> sum5_1(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pa3 = (double*) ppa[3];
double* pa4 = (double*) ppa[4];
double* pa5 = (double*) ppa[5];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub (a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub (a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub (a5Re, b0, a5Im_b0Swap);
SIMD<double> b1 (pb1 + i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
SIMD<double> a3Im_b1Swap = a3Im * b1Swap;
SIMD<double> a4Im_b1Swap = a4Im * b1Swap;
SIMD<double> a5Im_b1Swap = a5Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub (a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub (a2Re, b1, a2Im_b1Swap);
sum3_1 += FMAddSub (a3Re, b1, a3Im_b1Swap);
sum4_1 += FMAddSub (a4Re, b1, a4Im_b1Swap);
sum5_1 += FMAddSub (a5Re, b1, a5Im_b1Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i, mask);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i, mask);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i, mask);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub(a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub(a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub(a5Re, b0, a5Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
SIMD<double> a3Im_b1Swap = a3Im * b1Swap;
SIMD<double> a4Im_b1Swap = a4Im * b1Swap;
SIMD<double> a5Im_b1Swap = a5Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub(a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub(a2Re, b1, a2Im_b1Swap);
sum3_1 += FMAddSub(a3Re, b1, a3Im_b1Swap);
sum4_1 += FMAddSub(a4Re, b1, a4Im_b1Swap);
sum5_1 += FMAddSub(a5Re, b1, a5Im_b1Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_1);
pc[1+1*dc].real(hsum[0]);
pc[1+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_1);
pc[1+2*dc].real(hsum[0]);
pc[1+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_0);
pc[0+3*dc].real(hsum[0]);
pc[0+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_1);
pc[1+3*dc].real(hsum[0]);
pc[1+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_0);
pc[0+4*dc].real(hsum[0]);
pc[0+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_1);
pc[1+4*dc].real(hsum[0]);
pc[1+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_0);
pc[0+5*dc].real(hsum[0]);
pc[0+5*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_1);
pc[1+5*dc].real(hsum[0]);
pc[1+5*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<6, 1, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum1_0(0);
SIMD<double> sum2_0(0);
SIMD<double> sum3_0(0);
SIMD<double> sum4_0(0);
SIMD<double> sum5_0(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pa3 = (double*) ppa[3];
double* pa4 = (double*) ppa[4];
double* pa5 = (double*) ppa[5];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub (a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub (a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub (a5Re, b0, a5Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i, mask);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i, mask);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i, mask);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub(a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub(a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub(a5Re, b0, a5Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_0);
pc[0+3*dc].real(hsum[0]);
pc[0+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_0);
pc[0+4*dc].real(hsum[0]);
pc[0+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_0);
pc[0+5*dc].real(hsum[0]);
pc[0+5*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<3, 2, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum1_0(0);
SIMD<double> sum1_1(0);
SIMD<double> sum2_0(0);
SIMD<double> sum2_1(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
SIMD<double> b1 (pb1 + i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub (a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub (a2Re, b1, a2Im_b1Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub(a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub(a2Re, b1, a2Im_b1Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_1);
pc[1+1*dc].real(hsum[0]);
pc[1+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_1);
pc[1+2*dc].real(hsum[0]);
pc[1+2*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<3, 1, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum1_0(0);
SIMD<double> sum2_0(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 8, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum0_2(0);
SIMD<double> sum0_3(0);
SIMD<double> sum0_4(0);
SIMD<double> sum0_5(0);
SIMD<double> sum0_6(0);
SIMD<double> sum0_7(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
double* pb4 = (double*) ppb[4];
double* pb5 = (double*) ppb[5];
double* pb6 = (double*) ppb[6];
double* pb7 = (double*) ppb[7];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub (a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub (a0Re, b3, a0Im_b3Swap);
SIMD<double> b4 (pb4 + i);
SIMD<double> b4Swap = SwapPairs(b4);
SIMD<double> a0Im_b4Swap = a0Im * b4Swap;
sum0_4 += FMAddSub (a0Re, b4, a0Im_b4Swap);
SIMD<double> b5 (pb5 + i);
SIMD<double> b5Swap = SwapPairs(b5);
SIMD<double> a0Im_b5Swap = a0Im * b5Swap;
sum0_5 += FMAddSub (a0Re, b5, a0Im_b5Swap);
SIMD<double> b6 (pb6 + i);
SIMD<double> b6Swap = SwapPairs(b6);
SIMD<double> a0Im_b6Swap = a0Im * b6Swap;
sum0_6 += FMAddSub (a0Re, b6, a0Im_b6Swap);
SIMD<double> b7 (pb7 + i);
SIMD<double> b7Swap = SwapPairs(b7);
SIMD<double> a0Im_b7Swap = a0Im * b7Swap;
sum0_7 += FMAddSub (a0Re, b7, a0Im_b7Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i, mask);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub(a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i, mask);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub(a0Re, b3, a0Im_b3Swap);
SIMD<double> b4 (pb4 + i, mask);
SIMD<double> b4Swap = SwapPairs(b4);
SIMD<double> a0Im_b4Swap = a0Im * b4Swap;
sum0_4 += FMAddSub(a0Re, b4, a0Im_b4Swap);
SIMD<double> b5 (pb5 + i, mask);
SIMD<double> b5Swap = SwapPairs(b5);
SIMD<double> a0Im_b5Swap = a0Im * b5Swap;
sum0_5 += FMAddSub(a0Re, b5, a0Im_b5Swap);
SIMD<double> b6 (pb6 + i, mask);
SIMD<double> b6Swap = SwapPairs(b6);
SIMD<double> a0Im_b6Swap = a0Im * b6Swap;
sum0_6 += FMAddSub(a0Re, b6, a0Im_b6Swap);
SIMD<double> b7 (pb7 + i, mask);
SIMD<double> b7Swap = SwapPairs(b7);
SIMD<double> a0Im_b7Swap = a0Im * b7Swap;
sum0_7 += FMAddSub(a0Re, b7, a0Im_b7Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_2);
pc[2+0*dc].real(hsum[0]);
pc[2+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_3);
pc[3+0*dc].real(hsum[0]);
pc[3+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_4);
pc[4+0*dc].real(hsum[0]);
pc[4+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_5);
pc[5+0*dc].real(hsum[0]);
pc[5+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_6);
pc[6+0*dc].real(hsum[0]);
pc[6+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_7);
pc[7+0*dc].real(hsum[0]);
pc[7+0*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 4, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum0_2(0);
SIMD<double> sum0_3(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub (a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub (a0Re, b3, a0Im_b3Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i, mask);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub(a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i, mask);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub(a0Re, b3, a0Im_b3Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_2);
pc[2+0*dc].real(hsum[0]);
pc[2+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_3);
pc[3+0*dc].real(hsum[0]);
pc[3+0*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 1, 0>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<6, 2, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum1_0(0);
SIMD<double> sum1_1(0);
SIMD<double> sum2_0(0);
SIMD<double> sum2_1(0);
SIMD<double> sum3_0(0);
SIMD<double> sum3_1(0);
SIMD<double> sum4_0(0);
SIMD<double> sum4_1(0);
SIMD<double> sum5_0(0);
SIMD<double> sum5_1(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pa3 = (double*) ppa[3];
double* pa4 = (double*) ppa[4];
double* pa5 = (double*) ppa[5];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub (a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub (a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub (a5Re, b0, a5Im_b0Swap);
SIMD<double> b1 (pb1 + i);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
SIMD<double> a3Im_b1Swap = a3Im * b1Swap;
SIMD<double> a4Im_b1Swap = a4Im * b1Swap;
SIMD<double> a5Im_b1Swap = a5Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub (a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub (a2Re, b1, a2Im_b1Swap);
sum3_1 += FMAddSub (a3Re, b1, a3Im_b1Swap);
sum4_1 += FMAddSub (a4Re, b1, a4Im_b1Swap);
sum5_1 += FMAddSub (a5Re, b1, a5Im_b1Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i, mask);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i, mask);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i, mask);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub(a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub(a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub(a5Re, b0, a5Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
SIMD<double> a3Im_b1Swap = a3Im * b1Swap;
SIMD<double> a4Im_b1Swap = a4Im * b1Swap;
SIMD<double> a5Im_b1Swap = a5Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub(a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub(a2Re, b1, a2Im_b1Swap);
sum3_1 += FMAddSub(a3Re, b1, a3Im_b1Swap);
sum4_1 += FMAddSub(a4Re, b1, a4Im_b1Swap);
sum5_1 += FMAddSub(a5Re, b1, a5Im_b1Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_1);
pc[1+1*dc].real(hsum[0]);
pc[1+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_1);
pc[1+2*dc].real(hsum[0]);
pc[1+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_0);
pc[0+3*dc].real(hsum[0]);
pc[0+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_1);
pc[1+3*dc].real(hsum[0]);
pc[1+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_0);
pc[0+4*dc].real(hsum[0]);
pc[0+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_1);
pc[1+4*dc].real(hsum[0]);
pc[1+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_0);
pc[0+5*dc].real(hsum[0]);
pc[0+5*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_1);
pc[1+5*dc].real(hsum[0]);
pc[1+5*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<6, 1, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum1_0(0);
SIMD<double> sum2_0(0);
SIMD<double> sum3_0(0);
SIMD<double> sum4_0(0);
SIMD<double> sum5_0(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pa3 = (double*) ppa[3];
double* pa4 = (double*) ppa[4];
double* pa5 = (double*) ppa[5];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub (a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub (a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub (a5Re, b0, a5Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> a3(pa3 + i, mask);
auto [a3Re, a3Im] = Unpack(a3, a3);
SIMD<double> a4(pa4 + i, mask);
auto [a4Re, a4Im] = Unpack(a4, a4);
SIMD<double> a5(pa5 + i, mask);
auto [a5Re, a5Im] = Unpack(a5, a5);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
SIMD<double> a3Im_b0Swap = a3Im * b0Swap;
SIMD<double> a4Im_b0Swap = a4Im * b0Swap;
SIMD<double> a5Im_b0Swap = a5Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
sum3_0 += FMAddSub(a3Re, b0, a3Im_b0Swap);
sum4_0 += FMAddSub(a4Re, b0, a4Im_b0Swap);
sum5_0 += FMAddSub(a5Re, b0, a5Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum3_0);
pc[0+3*dc].real(hsum[0]);
pc[0+3*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum4_0);
pc[0+4*dc].real(hsum[0]);
pc[0+4*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum5_0);
pc[0+5*dc].real(hsum[0]);
pc[0+5*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<3, 2, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum1_0(0);
SIMD<double> sum1_1(0);
SIMD<double> sum2_0(0);
SIMD<double> sum2_1(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
SIMD<double> b1 (pb1 + i);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub (a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub (a2Re, b1, a2Im_b1Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
SIMD<double> a1Im_b1Swap = a1Im * b1Swap;
SIMD<double> a2Im_b1Swap = a2Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
sum1_1 += FMAddSub(a1Re, b1, a1Im_b1Swap);
sum2_1 += FMAddSub(a2Re, b1, a2Im_b1Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_1);
pc[1+1*dc].real(hsum[0]);
pc[1+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_1);
pc[1+2*dc].real(hsum[0]);
pc[1+2*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<3, 1, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum1_0(0);
SIMD<double> sum2_0(0);
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub (a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub (a2Re, b0, a2Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> a1(pa1 + i, mask);
auto [a1Re, a1Im] = Unpack(a1, a1);
SIMD<double> a2(pa2 + i, mask);
auto [a2Re, a2Im] = Unpack(a2, a2);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
SIMD<double> a1Im_b0Swap = a1Im * b0Swap;
SIMD<double> a2Im_b0Swap = a2Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
sum1_0 += FMAddSub(a1Re, b0, a1Im_b0Swap);
sum2_0 += FMAddSub(a2Re, b0, a2Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum1_0);
pc[0+1*dc].real(hsum[0]);
pc[0+1*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum2_0);
pc[0+2*dc].real(hsum[0]);
pc[0+2*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 8, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum0_2(0);
SIMD<double> sum0_3(0);
SIMD<double> sum0_4(0);
SIMD<double> sum0_5(0);
SIMD<double> sum0_6(0);
SIMD<double> sum0_7(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
double* pb4 = (double*) ppb[4];
double* pb5 = (double*) ppb[5];
double* pb6 = (double*) ppb[6];
double* pb7 = (double*) ppb[7];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i);
b2 *= conj;
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub (a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i);
b3 *= conj;
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub (a0Re, b3, a0Im_b3Swap);
SIMD<double> b4 (pb4 + i);
b4 *= conj;
SIMD<double> b4Swap = SwapPairs(b4);
SIMD<double> a0Im_b4Swap = a0Im * b4Swap;
sum0_4 += FMAddSub (a0Re, b4, a0Im_b4Swap);
SIMD<double> b5 (pb5 + i);
b5 *= conj;
SIMD<double> b5Swap = SwapPairs(b5);
SIMD<double> a0Im_b5Swap = a0Im * b5Swap;
sum0_5 += FMAddSub (a0Re, b5, a0Im_b5Swap);
SIMD<double> b6 (pb6 + i);
b6 *= conj;
SIMD<double> b6Swap = SwapPairs(b6);
SIMD<double> a0Im_b6Swap = a0Im * b6Swap;
sum0_6 += FMAddSub (a0Re, b6, a0Im_b6Swap);
SIMD<double> b7 (pb7 + i);
b7 *= conj;
SIMD<double> b7Swap = SwapPairs(b7);
SIMD<double> a0Im_b7Swap = a0Im * b7Swap;
sum0_7 += FMAddSub (a0Re, b7, a0Im_b7Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i, mask);
b2 *= conj;
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub(a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i, mask);
b3 *= conj;
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub(a0Re, b3, a0Im_b3Swap);
SIMD<double> b4 (pb4 + i, mask);
b4 *= conj;
SIMD<double> b4Swap = SwapPairs(b4);
SIMD<double> a0Im_b4Swap = a0Im * b4Swap;
sum0_4 += FMAddSub(a0Re, b4, a0Im_b4Swap);
SIMD<double> b5 (pb5 + i, mask);
b5 *= conj;
SIMD<double> b5Swap = SwapPairs(b5);
SIMD<double> a0Im_b5Swap = a0Im * b5Swap;
sum0_5 += FMAddSub(a0Re, b5, a0Im_b5Swap);
SIMD<double> b6 (pb6 + i, mask);
b6 *= conj;
SIMD<double> b6Swap = SwapPairs(b6);
SIMD<double> a0Im_b6Swap = a0Im * b6Swap;
sum0_6 += FMAddSub(a0Re, b6, a0Im_b6Swap);
SIMD<double> b7 (pb7 + i, mask);
b7 *= conj;
SIMD<double> b7Swap = SwapPairs(b7);
SIMD<double> a0Im_b7Swap = a0Im * b7Swap;
sum0_7 += FMAddSub(a0Re, b7, a0Im_b7Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_2);
pc[2+0*dc].real(hsum[0]);
pc[2+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_3);
pc[3+0*dc].real(hsum[0]);
pc[3+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_4);
pc[4+0*dc].real(hsum[0]);
pc[4+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_5);
pc[5+0*dc].real(hsum[0]);
pc[5+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_6);
pc[6+0*dc].real(hsum[0]);
pc[6+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_7);
pc[7+0*dc].real(hsum[0]);
pc[7+0*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 4, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
SIMD<double> sum0_1(0);
SIMD<double> sum0_2(0);
SIMD<double> sum0_3(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub (a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i);
b2 *= conj;
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub (a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i);
b3 *= conj;
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub (a0Re, b3, a0Im_b3Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
SIMD<double> b1 (pb1 + i, mask);
b1 *= conj;
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> a0Im_b1Swap = a0Im * b1Swap;
sum0_1 += FMAddSub(a0Re, b1, a0Im_b1Swap);
SIMD<double> b2 (pb2 + i, mask);
b2 *= conj;
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> a0Im_b2Swap = a0Im * b2Swap;
sum0_2 += FMAddSub(a0Re, b2, a0Im_b2Swap);
SIMD<double> b3 (pb3 + i, mask);
b3 *= conj;
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> a0Im_b3Swap = a0Im * b3Swap;
sum0_3 += FMAddSub(a0Re, b3, a0Im_b3Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_1);
pc[1+0*dc].real(hsum[0]);
pc[1+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_2);
pc[2+0*dc].real(hsum[0]);
pc[2+0*dc].imag(hsum[1]); }
{ SIMD<double,2> hsum = HSum128(sum0_3);
pc[3+0*dc].real(hsum[0]);
pc[3+0*dc].imag(hsum[1]); }
}
template <> INLINE void MultiVecScalC<1, 1, 1>
    (size_t n,
     Complex** ppa,
     Complex** ppb,
     Complex* pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> conj([](int i) { return (i%2==0) ? 1. : -1.; });
SIMD<double> sum0_0(0);
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for (; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0 + i);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub (a0Re, b0, a0Im_b0Swap);
}
int r = (2*n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0 + i, mask);
auto [a0Re, a0Im] = Unpack(a0, a0);
SIMD<double> b0 (pb0 + i, mask);
b0 *= conj;
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> a0Im_b0Swap = a0Im * b0Swap;
sum0_0 += FMAddSub(a0Re, b0, a0Im_b0Swap);
}
{ SIMD<double,2> hsum = HSum128(sum0_0);
pc[0+0*dc].real(hsum[0]);
pc[0+0*dc].imag(hsum[1]); }
}
template <size_t H, size_t W> inline void MultiScaleAdd
    (size_t n,
     double ** pa, 
     double ** pb, 
     double * pc, size_t dc);
template <> INLINE void MultiScaleAdd<6, 6>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double c0_2 = pc[0+2*dc];
double c0_3 = pc[0+3*dc];
double c0_4 = pc[0+4*dc];
double c0_5 = pc[0+5*dc];
double c1_0 = pc[1+0*dc];
double c1_1 = pc[1+1*dc];
double c1_2 = pc[1+2*dc];
double c1_3 = pc[1+3*dc];
double c1_4 = pc[1+4*dc];
double c1_5 = pc[1+5*dc];
double c2_0 = pc[2+0*dc];
double c2_1 = pc[2+1*dc];
double c2_2 = pc[2+2*dc];
double c2_3 = pc[2+3*dc];
double c2_4 = pc[2+4*dc];
double c2_5 = pc[2+5*dc];
double c3_0 = pc[3+0*dc];
double c3_1 = pc[3+1*dc];
double c3_2 = pc[3+2*dc];
double c3_3 = pc[3+3*dc];
double c3_4 = pc[3+4*dc];
double c3_5 = pc[3+5*dc];
double c4_0 = pc[4+0*dc];
double c4_1 = pc[4+1*dc];
double c4_2 = pc[4+2*dc];
double c4_3 = pc[4+3*dc];
double c4_4 = pc[4+4*dc];
double c4_5 = pc[4+5*dc];
double c5_0 = pc[5+0*dc];
double c5_1 = pc[5+1*dc];
double c5_2 = pc[5+2*dc];
double c5_3 = pc[5+3*dc];
double c5_4 = pc[5+4*dc];
double c5_5 = pc[5+5*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
double* pb4 = ppb[4];
double* pb5 = ppb[5];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a2 += c2_1 * b1;
a3 += c3_1 * b1;
a4 += c4_1 * b1;
a5 += c5_1 * b1;
SIMD<double> b2(pb2+i);
a0 += c0_2 * b2;
a1 += c1_2 * b2;
a2 += c2_2 * b2;
a3 += c3_2 * b2;
a4 += c4_2 * b2;
a5 += c5_2 * b2;
SIMD<double> b3(pb3+i);
a0 += c0_3 * b3;
a1 += c1_3 * b3;
a2 += c2_3 * b3;
a3 += c3_3 * b3;
a4 += c4_3 * b3;
a5 += c5_3 * b3;
SIMD<double> b4(pb4+i);
a0 += c0_4 * b4;
a1 += c1_4 * b4;
a2 += c2_4 * b4;
a3 += c3_4 * b4;
a4 += c4_4 * b4;
a5 += c5_4 * b4;
SIMD<double> b5(pb5+i);
a0 += c0_5 * b5;
a1 += c1_5 * b5;
a2 += c2_5 * b5;
a3 += c3_5 * b5;
a4 += c4_5 * b5;
a5 += c5_5 * b5;
a0.Store(pa0+i);
a1.Store(pa1+i);
a2.Store(pa2+i);
a3.Store(pa3+i);
a4.Store(pa4+i);
a5.Store(pa5+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a2 += c2_1 * b1;
a3 += c3_1 * b1;
a4 += c4_1 * b1;
a5 += c5_1 * b1;
SIMD<double> b2(pb2+i, mask);
a0 += c0_2 * b2;
a1 += c1_2 * b2;
a2 += c2_2 * b2;
a3 += c3_2 * b2;
a4 += c4_2 * b2;
a5 += c5_2 * b2;
SIMD<double> b3(pb3+i, mask);
a0 += c0_3 * b3;
a1 += c1_3 * b3;
a2 += c2_3 * b3;
a3 += c3_3 * b3;
a4 += c4_3 * b3;
a5 += c5_3 * b3;
SIMD<double> b4(pb4+i, mask);
a0 += c0_4 * b4;
a1 += c1_4 * b4;
a2 += c2_4 * b4;
a3 += c3_4 * b4;
a4 += c4_4 * b4;
a5 += c5_4 * b4;
SIMD<double> b5(pb5+i, mask);
a0 += c0_5 * b5;
a1 += c1_5 * b5;
a2 += c2_5 * b5;
a3 += c3_5 * b5;
a4 += c4_5 * b5;
a5 += c5_5 * b5;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
a2.Store(pa2+i, mask);
a3.Store(pa3+i, mask);
a4.Store(pa4+i, mask);
a5.Store(pa5+i, mask);
}
}
template <> INLINE void MultiScaleAdd<2, 6>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double c0_2 = pc[0+2*dc];
double c0_3 = pc[0+3*dc];
double c0_4 = pc[0+4*dc];
double c0_5 = pc[0+5*dc];
double c1_0 = pc[1+0*dc];
double c1_1 = pc[1+1*dc];
double c1_2 = pc[1+2*dc];
double c1_3 = pc[1+3*dc];
double c1_4 = pc[1+4*dc];
double c1_5 = pc[1+5*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
double* pb4 = ppb[4];
double* pb5 = ppb[5];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
SIMD<double> b2(pb2+i);
a0 += c0_2 * b2;
a1 += c1_2 * b2;
SIMD<double> b3(pb3+i);
a0 += c0_3 * b3;
a1 += c1_3 * b3;
SIMD<double> b4(pb4+i);
a0 += c0_4 * b4;
a1 += c1_4 * b4;
SIMD<double> b5(pb5+i);
a0 += c0_5 * b5;
a1 += c1_5 * b5;
a0.Store(pa0+i);
a1.Store(pa1+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
SIMD<double> b2(pb2+i, mask);
a0 += c0_2 * b2;
a1 += c1_2 * b2;
SIMD<double> b3(pb3+i, mask);
a0 += c0_3 * b3;
a1 += c1_3 * b3;
SIMD<double> b4(pb4+i, mask);
a0 += c0_4 * b4;
a1 += c1_4 * b4;
SIMD<double> b5(pb5+i, mask);
a0 += c0_5 * b5;
a1 += c1_5 * b5;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
}
}
template <> INLINE void MultiScaleAdd<1, 6>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double c0_2 = pc[0+2*dc];
double c0_3 = pc[0+3*dc];
double c0_4 = pc[0+4*dc];
double c0_5 = pc[0+5*dc];
double* pa0 = ppa[0];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
double* pb2 = ppb[2];
double* pb3 = ppb[3];
double* pb4 = ppb[4];
double* pb5 = ppb[5];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
SIMD<double> b2(pb2+i);
a0 += c0_2 * b2;
SIMD<double> b3(pb3+i);
a0 += c0_3 * b3;
SIMD<double> b4(pb4+i);
a0 += c0_4 * b4;
SIMD<double> b5(pb5+i);
a0 += c0_5 * b5;
a0.Store(pa0+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
SIMD<double> b2(pb2+i, mask);
a0 += c0_2 * b2;
SIMD<double> b3(pb3+i, mask);
a0 += c0_3 * b3;
SIMD<double> b4(pb4+i, mask);
a0 += c0_4 * b4;
SIMD<double> b5(pb5+i, mask);
a0 += c0_5 * b5;
a0.Store(pa0+i, mask);
}
}
template <> INLINE void MultiScaleAdd<6, 2>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double c1_0 = pc[1+0*dc];
double c1_1 = pc[1+1*dc];
double c2_0 = pc[2+0*dc];
double c2_1 = pc[2+1*dc];
double c3_0 = pc[3+0*dc];
double c3_1 = pc[3+1*dc];
double c4_0 = pc[4+0*dc];
double c4_1 = pc[4+1*dc];
double c5_0 = pc[5+0*dc];
double c5_1 = pc[5+1*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a2 += c2_1 * b1;
a3 += c3_1 * b1;
a4 += c4_1 * b1;
a5 += c5_1 * b1;
a0.Store(pa0+i);
a1.Store(pa1+i);
a2.Store(pa2+i);
a3.Store(pa3+i);
a4.Store(pa4+i);
a5.Store(pa5+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a2 += c2_1 * b1;
a3 += c3_1 * b1;
a4 += c4_1 * b1;
a5 += c5_1 * b1;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
a2.Store(pa2+i, mask);
a3.Store(pa3+i, mask);
a4.Store(pa4+i, mask);
a5.Store(pa5+i, mask);
}
}
template <> INLINE void MultiScaleAdd<2, 2>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double c1_0 = pc[1+0*dc];
double c1_1 = pc[1+1*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a0.Store(pa0+i);
a1.Store(pa1+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
a1 += c1_1 * b1;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
}
}
template <> INLINE void MultiScaleAdd<1, 2>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c0_1 = pc[0+1*dc];
double* pa0 = ppa[0];
double* pb0 = ppb[0];
double* pb1 = ppb[1];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
SIMD<double> b1(pb1+i);
a0 += c0_1 * b1;
a0.Store(pa0+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
SIMD<double> b1(pb1+i, mask);
a0 += c0_1 * b1;
a0.Store(pa0+i, mask);
}
}
template <> INLINE void MultiScaleAdd<6, 1>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c1_0 = pc[1+0*dc];
double c2_0 = pc[2+0*dc];
double c3_0 = pc[3+0*dc];
double c4_0 = pc[4+0*dc];
double c5_0 = pc[5+0*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pa2 = ppa[2];
double* pa3 = ppa[3];
double* pa4 = ppa[4];
double* pa5 = ppa[5];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> a3(pa3+i);
SIMD<double> a4(pa4+i);
SIMD<double> a5(pa5+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
a0.Store(pa0+i);
a1.Store(pa1+i);
a2.Store(pa2+i);
a3.Store(pa3+i);
a4.Store(pa4+i);
a5.Store(pa5+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> a3(pa3+i, mask);
SIMD<double> a4(pa4+i, mask);
SIMD<double> a5(pa5+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a2 += c2_0 * b0;
a3 += c3_0 * b0;
a4 += c4_0 * b0;
a5 += c5_0 * b0;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
a2.Store(pa2+i, mask);
a3.Store(pa3+i, mask);
a4.Store(pa4+i, mask);
a5.Store(pa5+i, mask);
}
}
template <> INLINE void MultiScaleAdd<2, 1>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double c1_0 = pc[1+0*dc];
double* pa0 = ppa[0];
double* pa1 = ppa[1];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a0.Store(pa0+i);
a1.Store(pa1+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a1 += c1_0 * b0;
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
}
}
template <> INLINE void MultiScaleAdd<1, 1>
    (size_t n,
     double ** ppa, 
     double ** ppb, 
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double c0_0 = pc[0+0*dc];
double* pa0 = ppa[0];
double* pb0 = ppb[0];
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
a0 += c0_0 * b0;
a0.Store(pa0+i);
}
size_t r = n % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
a0 += c0_0 * b0;
a0.Store(pa0+i, mask);
}
}
template <size_t H, size_t W> inline void MultiScaleAddC
    (size_t n,
     Complex ** pa, 
     Complex ** pb, 
     Complex * pc, size_t dc);
template <> INLINE void MultiScaleAddC<3, 4>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
SIMD<double> c0_1Re = pc[0+1*dc].real();
SIMD<double> c0_1Im = pc[0+1*dc].imag();
SIMD<double> c0_2Re = pc[0+2*dc].real();
SIMD<double> c0_2Im = pc[0+2*dc].imag();
SIMD<double> c0_3Re = pc[0+3*dc].real();
SIMD<double> c0_3Im = pc[0+3*dc].imag();
SIMD<double> c1_0Re = pc[1+0*dc].real();
SIMD<double> c1_0Im = pc[1+0*dc].imag();
SIMD<double> c1_1Re = pc[1+1*dc].real();
SIMD<double> c1_1Im = pc[1+1*dc].imag();
SIMD<double> c1_2Re = pc[1+2*dc].real();
SIMD<double> c1_2Im = pc[1+2*dc].imag();
SIMD<double> c1_3Re = pc[1+3*dc].real();
SIMD<double> c1_3Im = pc[1+3*dc].imag();
SIMD<double> c2_0Re = pc[2+0*dc].real();
SIMD<double> c2_0Im = pc[2+0*dc].imag();
SIMD<double> c2_1Re = pc[2+1*dc].real();
SIMD<double> c2_1Im = pc[2+1*dc].imag();
SIMD<double> c2_2Re = pc[2+2*dc].real();
SIMD<double> c2_2Im = pc[2+2*dc].imag();
SIMD<double> c2_3Re = pc[2+3*dc].real();
SIMD<double> c2_3Im = pc[2+3*dc].imag();
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub (c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> c2_0Im_b0Swap = c2_0Im * b0Swap;
a2 += FMAddSub (c2_0Re, b0, c2_0Im_b0Swap);
SIMD<double> b1(pb1+i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub (c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> c1_1Im_b1Swap = c1_1Im * b1Swap;
a1 += FMAddSub (c1_1Re, b1, c1_1Im_b1Swap);
SIMD<double> c2_1Im_b1Swap = c2_1Im * b1Swap;
a2 += FMAddSub (c2_1Re, b1, c2_1Im_b1Swap);
SIMD<double> b2(pb2+i);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub (c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> c1_2Im_b2Swap = c1_2Im * b2Swap;
a1 += FMAddSub (c1_2Re, b2, c1_2Im_b2Swap);
SIMD<double> c2_2Im_b2Swap = c2_2Im * b2Swap;
a2 += FMAddSub (c2_2Re, b2, c2_2Im_b2Swap);
SIMD<double> b3(pb3+i);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub (c0_3Re, b3, c0_3Im_b3Swap);
SIMD<double> c1_3Im_b3Swap = c1_3Im * b3Swap;
a1 += FMAddSub (c1_3Re, b3, c1_3Im_b3Swap);
SIMD<double> c2_3Im_b3Swap = c2_3Im * b3Swap;
a2 += FMAddSub (c2_3Re, b3, c2_3Im_b3Swap);
a0.Store(pa0+i);
a1.Store(pa1+i);
a2.Store(pa2+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub(c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> c2_0Im_b0Swap = c2_0Im * b0Swap;
a2 += FMAddSub(c2_0Re, b0, c2_0Im_b0Swap);
SIMD<double> b1(pb1+i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub(c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> c1_1Im_b1Swap = c1_1Im * b1Swap;
a1 += FMAddSub(c1_1Re, b1, c1_1Im_b1Swap);
SIMD<double> c2_1Im_b1Swap = c2_1Im * b1Swap;
a2 += FMAddSub(c2_1Re, b1, c2_1Im_b1Swap);
SIMD<double> b2(pb2+i, mask);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub(c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> c1_2Im_b2Swap = c1_2Im * b2Swap;
a1 += FMAddSub(c1_2Re, b2, c1_2Im_b2Swap);
SIMD<double> c2_2Im_b2Swap = c2_2Im * b2Swap;
a2 += FMAddSub(c2_2Re, b2, c2_2Im_b2Swap);
SIMD<double> b3(pb3+i, mask);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub(c0_3Re, b3, c0_3Im_b3Swap);
SIMD<double> c1_3Im_b3Swap = c1_3Im * b3Swap;
a1 += FMAddSub(c1_3Re, b3, c1_3Im_b3Swap);
SIMD<double> c2_3Im_b3Swap = c2_3Im * b3Swap;
a2 += FMAddSub(c2_3Re, b3, c2_3Im_b3Swap);
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
a2.Store(pa2+i, mask);
}
}
template <> INLINE void MultiScaleAddC<2, 4>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
SIMD<double> c0_1Re = pc[0+1*dc].real();
SIMD<double> c0_1Im = pc[0+1*dc].imag();
SIMD<double> c0_2Re = pc[0+2*dc].real();
SIMD<double> c0_2Im = pc[0+2*dc].imag();
SIMD<double> c0_3Re = pc[0+3*dc].real();
SIMD<double> c0_3Im = pc[0+3*dc].imag();
SIMD<double> c1_0Re = pc[1+0*dc].real();
SIMD<double> c1_0Im = pc[1+0*dc].imag();
SIMD<double> c1_1Re = pc[1+1*dc].real();
SIMD<double> c1_1Im = pc[1+1*dc].imag();
SIMD<double> c1_2Re = pc[1+2*dc].real();
SIMD<double> c1_2Im = pc[1+2*dc].imag();
SIMD<double> c1_3Re = pc[1+3*dc].real();
SIMD<double> c1_3Im = pc[1+3*dc].imag();
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub (c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> b1(pb1+i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub (c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> c1_1Im_b1Swap = c1_1Im * b1Swap;
a1 += FMAddSub (c1_1Re, b1, c1_1Im_b1Swap);
SIMD<double> b2(pb2+i);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub (c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> c1_2Im_b2Swap = c1_2Im * b2Swap;
a1 += FMAddSub (c1_2Re, b2, c1_2Im_b2Swap);
SIMD<double> b3(pb3+i);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub (c0_3Re, b3, c0_3Im_b3Swap);
SIMD<double> c1_3Im_b3Swap = c1_3Im * b3Swap;
a1 += FMAddSub (c1_3Re, b3, c1_3Im_b3Swap);
a0.Store(pa0+i);
a1.Store(pa1+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub(c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> b1(pb1+i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub(c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> c1_1Im_b1Swap = c1_1Im * b1Swap;
a1 += FMAddSub(c1_1Re, b1, c1_1Im_b1Swap);
SIMD<double> b2(pb2+i, mask);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub(c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> c1_2Im_b2Swap = c1_2Im * b2Swap;
a1 += FMAddSub(c1_2Re, b2, c1_2Im_b2Swap);
SIMD<double> b3(pb3+i, mask);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub(c0_3Re, b3, c0_3Im_b3Swap);
SIMD<double> c1_3Im_b3Swap = c1_3Im * b3Swap;
a1 += FMAddSub(c1_3Re, b3, c1_3Im_b3Swap);
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
}
}
template <> INLINE void MultiScaleAddC<1, 4>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
SIMD<double> c0_1Re = pc[0+1*dc].real();
SIMD<double> c0_1Im = pc[0+1*dc].imag();
SIMD<double> c0_2Re = pc[0+2*dc].real();
SIMD<double> c0_2Im = pc[0+2*dc].imag();
SIMD<double> c0_3Re = pc[0+3*dc].real();
SIMD<double> c0_3Im = pc[0+3*dc].imag();
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
double* pb1 = (double*) ppb[1];
double* pb2 = (double*) ppb[2];
double* pb3 = (double*) ppb[3];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> b1(pb1+i);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub (c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> b2(pb2+i);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub (c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> b3(pb3+i);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub (c0_3Re, b3, c0_3Im_b3Swap);
a0.Store(pa0+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> b1(pb1+i, mask);
SIMD<double> b1Swap = SwapPairs(b1);
SIMD<double> c0_1Im_b1Swap = c0_1Im * b1Swap;
a0 += FMAddSub(c0_1Re, b1, c0_1Im_b1Swap);
SIMD<double> b2(pb2+i, mask);
SIMD<double> b2Swap = SwapPairs(b2);
SIMD<double> c0_2Im_b2Swap = c0_2Im * b2Swap;
a0 += FMAddSub(c0_2Re, b2, c0_2Im_b2Swap);
SIMD<double> b3(pb3+i, mask);
SIMD<double> b3Swap = SwapPairs(b3);
SIMD<double> c0_3Im_b3Swap = c0_3Im * b3Swap;
a0 += FMAddSub(c0_3Re, b3, c0_3Im_b3Swap);
a0.Store(pa0+i, mask);
}
}
template <> INLINE void MultiScaleAddC<3, 1>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
SIMD<double> c1_0Re = pc[1+0*dc].real();
SIMD<double> c1_0Im = pc[1+0*dc].imag();
SIMD<double> c2_0Re = pc[2+0*dc].real();
SIMD<double> c2_0Im = pc[2+0*dc].imag();
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pa2 = (double*) ppa[2];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> a2(pa2+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub (c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> c2_0Im_b0Swap = c2_0Im * b0Swap;
a2 += FMAddSub (c2_0Re, b0, c2_0Im_b0Swap);
a0.Store(pa0+i);
a1.Store(pa1+i);
a2.Store(pa2+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> a2(pa2+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub(c1_0Re, b0, c1_0Im_b0Swap);
SIMD<double> c2_0Im_b0Swap = c2_0Im * b0Swap;
a2 += FMAddSub(c2_0Re, b0, c2_0Im_b0Swap);
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
a2.Store(pa2+i, mask);
}
}
template <> INLINE void MultiScaleAddC<2, 1>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
SIMD<double> c1_0Re = pc[1+0*dc].real();
SIMD<double> c1_0Im = pc[1+0*dc].imag();
double* pa0 = (double*) ppa[0];
double* pa1 = (double*) ppa[1];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> a1(pa1+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub (c1_0Re, b0, c1_0Im_b0Swap);
a0.Store(pa0+i);
a1.Store(pa1+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> a1(pa1+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
SIMD<double> c1_0Im_b0Swap = c1_0Im * b0Swap;
a1 += FMAddSub(c1_0Re, b0, c1_0Im_b0Swap);
a0.Store(pa0+i, mask);
a1.Store(pa1+i, mask);
}
}
template <> INLINE void MultiScaleAddC<1, 1>
    (size_t n,
     Complex ** ppa, 
     Complex ** ppb, 
     Complex * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> c0_0Re = pc[0+0*dc].real();
SIMD<double> c0_0Im = pc[0+0*dc].imag();
double* pa0 = (double*) ppa[0];
double* pb0 = (double*) ppb[0];
size_t i = 0;
for ( ; i+SW <= 2*n; i+=SW) {
SIMD<double> a0(pa0+i);
SIMD<double> b0(pb0+i);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub (c0_0Re, b0, c0_0Im_b0Swap);
a0.Store(pa0+i);
}
size_t r = (2 * n) % SW;
if (r) {
SIMD<mask64> mask(r);
SIMD<double> a0(pa0+i, mask);
SIMD<double> b0(pb0+i, mask);
SIMD<double> b0Swap = SwapPairs(b0);
SIMD<double> c0_0Im_b0Swap = c0_0Im * b0Swap;
a0 += FMAddSub(c0_0Re, b0, c0_0Im_b0Swap);
a0.Store(pa0+i, mask);
}
}
template <size_t H, size_t W>
inline void MyScalTrans
(size_t n, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <> inline void MyScalTrans<1, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
}
template <> inline void MyScalTrans<2, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> a1(pa[1]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
FMAasm(b0,a1,sum10);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
FMAasm(b1,a1,sum11);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
FMAasm(b2,a1,sum12);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
FMAasm(b3,a1,sum13);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
}
template <> inline void MyScalTrans<3, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> a1(pa[1]);
SIMD<double> a2(pa[2]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
FMAasm(b0,a1,sum10);
FMAasm(b0,a2,sum20);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
FMAasm(b1,a1,sum11);
FMAasm(b1,a2,sum21);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
FMAasm(b2,a1,sum12);
FMAasm(b2,a2,sum22);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
FMAasm(b3,a1,sum13);
FMAasm(b3,a2,sum23);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
}
template <> inline void MyScalTrans<4, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> a1(pa[1]);
SIMD<double> a2(pa[2]);
SIMD<double> a3(pa[3]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
FMAasm(b0,a1,sum10);
FMAasm(b0,a2,sum20);
FMAasm(b0,a3,sum30);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
FMAasm(b1,a1,sum11);
FMAasm(b1,a2,sum21);
FMAasm(b1,a3,sum31);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
FMAasm(b2,a1,sum12);
FMAasm(b2,a2,sum22);
FMAasm(b2,a3,sum32);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
FMAasm(b3,a1,sum13);
FMAasm(b3,a2,sum23);
FMAasm(b3,a3,sum33);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
}
template <> inline void MyScalTrans<5, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> a1(pa[1]);
SIMD<double> a2(pa[2]);
SIMD<double> a3(pa[3]);
SIMD<double> a4(pa[4]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
FMAasm(b0,a1,sum10);
FMAasm(b0,a2,sum20);
FMAasm(b0,a3,sum30);
FMAasm(b0,a4,sum40);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
FMAasm(b1,a1,sum11);
FMAasm(b1,a2,sum21);
FMAasm(b1,a3,sum31);
FMAasm(b1,a4,sum41);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
FMAasm(b2,a1,sum12);
FMAasm(b2,a2,sum22);
FMAasm(b2,a3,sum32);
FMAasm(b2,a4,sum42);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
FMAasm(b3,a1,sum13);
FMAasm(b3,a2,sum23);
FMAasm(b3,a3,sum33);
FMAasm(b3,a4,sum43);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
}
template <> inline void MyScalTrans<6, 4>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * hpc = pc;
SIMD<double> sum00(pc+SW*0);
SIMD<double> sum01(pc+SW*1);
SIMD<double> sum02(pc+SW*2);
SIMD<double> sum03(pc+SW*3);
pc += dc;
SIMD<double> sum10(pc+SW*0);
SIMD<double> sum11(pc+SW*1);
SIMD<double> sum12(pc+SW*2);
SIMD<double> sum13(pc+SW*3);
pc += dc;
SIMD<double> sum20(pc+SW*0);
SIMD<double> sum21(pc+SW*1);
SIMD<double> sum22(pc+SW*2);
SIMD<double> sum23(pc+SW*3);
pc += dc;
SIMD<double> sum30(pc+SW*0);
SIMD<double> sum31(pc+SW*1);
SIMD<double> sum32(pc+SW*2);
SIMD<double> sum33(pc+SW*3);
pc += dc;
SIMD<double> sum40(pc+SW*0);
SIMD<double> sum41(pc+SW*1);
SIMD<double> sum42(pc+SW*2);
SIMD<double> sum43(pc+SW*3);
pc += dc;
SIMD<double> sum50(pc+SW*0);
SIMD<double> sum51(pc+SW*1);
SIMD<double> sum52(pc+SW*2);
SIMD<double> sum53(pc+SW*3);
pc += dc;
pc = hpc;
for (size_t i = 0; i < n; i++, pa += da, pb += db) {
SIMD<double> a0(pa[0]);
SIMD<double> a1(pa[1]);
SIMD<double> a2(pa[2]);
SIMD<double> a3(pa[3]);
SIMD<double> a4(pa[4]);
SIMD<double> a5(pa[5]);
SIMD<double> b0(pb+0*SW);
FMAasm(b0,a0,sum00);
FMAasm(b0,a1,sum10);
FMAasm(b0,a2,sum20);
FMAasm(b0,a3,sum30);
FMAasm(b0,a4,sum40);
FMAasm(b0,a5,sum50);
SIMD<double> b1(pb+1*SW);
FMAasm(b1,a0,sum01);
FMAasm(b1,a1,sum11);
FMAasm(b1,a2,sum21);
FMAasm(b1,a3,sum31);
FMAasm(b1,a4,sum41);
FMAasm(b1,a5,sum51);
SIMD<double> b2(pb+2*SW);
FMAasm(b2,a0,sum02);
FMAasm(b2,a1,sum12);
FMAasm(b2,a2,sum22);
FMAasm(b2,a3,sum32);
FMAasm(b2,a4,sum42);
FMAasm(b2,a5,sum52);
SIMD<double> b3(pb+3*SW);
FMAasm(b3,a0,sum03);
FMAasm(b3,a1,sum13);
FMAasm(b3,a2,sum23);
FMAasm(b3,a3,sum33);
FMAasm(b3,a4,sum43);
FMAasm(b3,a5,sum53);
}
sum00.Store(pc+SW*0);
sum01.Store(pc+SW*1);
sum02.Store(pc+SW*2);
sum03.Store(pc+SW*3);
pc += dc;
sum10.Store(pc+SW*0);
sum11.Store(pc+SW*1);
sum12.Store(pc+SW*2);
sum13.Store(pc+SW*3);
pc += dc;
sum20.Store(pc+SW*0);
sum21.Store(pc+SW*1);
sum22.Store(pc+SW*2);
sum23.Store(pc+SW*3);
pc += dc;
sum30.Store(pc+SW*0);
sum31.Store(pc+SW*1);
sum32.Store(pc+SW*2);
sum33.Store(pc+SW*3);
pc += dc;
sum40.Store(pc+SW*0);
sum41.Store(pc+SW*1);
sum42.Store(pc+SW*2);
sum43.Store(pc+SW*3);
pc += dc;
sum50.Store(pc+SW*0);
sum51.Store(pc+SW*1);
sum52.Store(pc+SW*2);
sum53.Store(pc+SW*3);
pc += dc;
}
template <size_t H, size_t W, OPERATION OP>
inline void MatKernelDaxpy
(size_t n, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <size_t H, size_t W, OPERATION OP>
inline void MatKernelDaxpy
(size_t n, double * pa, size_t da, SIMD<double> * pb, size_t db, SIMD<double> * pc, size_t dc);
template <> INLINE void MatKernelDaxpy<1, 0, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * pc0 = pc+0*dc;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 0, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * pc0 = pc+0*dc;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 0, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
double * pc0 = pc+0*dc;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 5, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 5, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 5, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 6, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 6, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 6, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 7, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 7, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 7, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 8, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 8, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 8, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FNMAnonasm(a07, b7, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 -= a07 * b7;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 9, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 9, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 9, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FNMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FNMAnonasm(a08, b8, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 -= a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 -= a08 * b8;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 10, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 10, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 10, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FNMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FNMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FNMAnonasm(a09, b9, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 -= a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 -= a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 -= a09 * b9;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 11, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FMAnonasm(a010, b10, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 += a010 * b10;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 11, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FMAnonasm(a010, b10, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 += a010 * b10;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 11, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FNMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FNMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FNMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FNMAnonasm(a010, b10, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 -= a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 -= a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 -= a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 -= a010 * b10;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 12, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
SIMD<double> a011(pa[0*da+11]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
double * pb11 = pb+11*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FMAnonasm(a010, b10, c0);
SIMD<double> b11(pb11+i);
FMAnonasm(a011, b11, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 += a010 * b10;
SIMD<double> b11(pb11+i, mask);
c0 += a011 * b11;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 12, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
SIMD<double> a011(pa[0*da+11]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
double * pb11 = pb+11*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FMAnonasm(a010, b10, c0);
SIMD<double> b11(pb11+i);
FMAnonasm(a011, b11, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 += a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 += a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 += a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 += a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 += a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 += a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 += a010 * b10;
SIMD<double> b11(pb11+i, mask);
c0 += a011 * b11;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<1, 12, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a04(pa[0*da+4]);
SIMD<double> a05(pa[0*da+5]);
SIMD<double> a06(pa[0*da+6]);
SIMD<double> a07(pa[0*da+7]);
SIMD<double> a08(pa[0*da+8]);
SIMD<double> a09(pa[0*da+9]);
SIMD<double> a010(pa[0*da+10]);
SIMD<double> a011(pa[0*da+11]);
double * pc0 = pc+0*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
double * pb4 = pb+4*db;
double * pb5 = pb+5*db;
double * pb6 = pb+6*db;
double * pb7 = pb+7*db;
double * pb8 = pb+8*db;
double * pb9 = pb+9*db;
double * pb10 = pb+10*db;
double * pb11 = pb+11*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
SIMD<double> b4(pb4+i);
FNMAnonasm(a04, b4, c0);
SIMD<double> b5(pb5+i);
FNMAnonasm(a05, b5, c0);
SIMD<double> b6(pb6+i);
FNMAnonasm(a06, b6, c0);
SIMD<double> b7(pb7+i);
FNMAnonasm(a07, b7, c0);
SIMD<double> b8(pb8+i);
FNMAnonasm(a08, b8, c0);
SIMD<double> b9(pb9+i);
FNMAnonasm(a09, b9, c0);
SIMD<double> b10(pb10+i);
FNMAnonasm(a010, b10, c0);
SIMD<double> b11(pb11+i);
FNMAnonasm(a011, b11, c0);
c0.Store(pc0+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
SIMD<double> b4(pb4+i, mask);
c0 -= a04 * b4;
SIMD<double> b5(pb5+i, mask);
c0 -= a05 * b5;
SIMD<double> b6(pb6+i, mask);
c0 -= a06 * b6;
SIMD<double> b7(pb7+i, mask);
c0 -= a07 * b7;
SIMD<double> b8(pb8+i, mask);
c0 -= a08 * b8;
SIMD<double> b9(pb9+i, mask);
c0 -= a09 * b9;
SIMD<double> b10(pb10+i, mask);
c0 -= a010 * b10;
SIMD<double> b11(pb11+i, mask);
c0 -= a011 * b11;
c0.Store(pc0+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
FNMAnonasm(a12, b2, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
c1 -= a12 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
FMAnonasm(a13, b3, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c1 += a13 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
FMAnonasm(a13, b3, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c1 += a13 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<2, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
FNMAnonasm(a12, b2, c1);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
FNMAnonasm(a13, b3, c1);
c0.Store(pc0+i);
c1.Store(pc1+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
c1 -= a12 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
c1 -= a13 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 1, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a20(pa[2*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 1, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a20(pa[2*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 1, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a20(pa[2*da+0]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
FNMAnonasm(a20, b0, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
c2 -= a20 * b0;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 2, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 2, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 2, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
FNMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
FNMAnonasm(a21, b1, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
c2 -= a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
c2 -= a21 * b1;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 3, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
FMAnonasm(a22, b2, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c2 += a22 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 3, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
FMAnonasm(a22, b2, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c2 += a22 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 3, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
FNMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
FNMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
FNMAnonasm(a12, b2, c1);
FNMAnonasm(a22, b2, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
c2 -= a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
c2 -= a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
c1 -= a12 * b2;
c2 -= a22 * b2;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 4, SET>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
SIMD<double> a23(pa[2*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
FMAnonasm(a22, b2, c2);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
FMAnonasm(a13, b3, c1);
FMAnonasm(a23, b3, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(0);
SIMD<double> c1(0);
SIMD<double> c2(0);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c2 += a22 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c1 += a13 * b3;
c2 += a23 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 4, ADD>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
SIMD<double> a23(pa[2*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FMAnonasm(a00, b0, c0);
FMAnonasm(a10, b0, c1);
FMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FMAnonasm(a01, b1, c0);
FMAnonasm(a11, b1, c1);
FMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FMAnonasm(a02, b2, c0);
FMAnonasm(a12, b2, c1);
FMAnonasm(a22, b2, c2);
SIMD<double> b3(pb3+i);
FMAnonasm(a03, b3, c0);
FMAnonasm(a13, b3, c1);
FMAnonasm(a23, b3, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 += a00 * b0;
c1 += a10 * b0;
c2 += a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 += a01 * b1;
c1 += a11 * b1;
c2 += a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 += a02 * b2;
c1 += a12 * b2;
c2 += a22 * b2;
SIMD<double> b3(pb3+i, mask);
c0 += a03 * b3;
c1 += a13 * b3;
c2 += a23 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
template <> INLINE void MatKernelDaxpy<3, 4, SUB>
    (size_t n,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
SIMD<double> a00(pa[0*da+0]);
SIMD<double> a01(pa[0*da+1]);
SIMD<double> a02(pa[0*da+2]);
SIMD<double> a03(pa[0*da+3]);
SIMD<double> a10(pa[1*da+0]);
SIMD<double> a11(pa[1*da+1]);
SIMD<double> a12(pa[1*da+2]);
SIMD<double> a13(pa[1*da+3]);
SIMD<double> a20(pa[2*da+0]);
SIMD<double> a21(pa[2*da+1]);
SIMD<double> a22(pa[2*da+2]);
SIMD<double> a23(pa[2*da+3]);
double * pc0 = pc+0*dc;
double * pc1 = pc+1*dc;
double * pc2 = pc+2*dc;
double * pb0 = pb+0*db;
double * pb1 = pb+1*db;
double * pb2 = pb+2*db;
double * pb3 = pb+3*db;
size_t i = 0;
for ( ; i+SW <= n; i+=SW) {
SIMD<double> c0(pc0+i);
SIMD<double> c1(pc1+i);
SIMD<double> c2(pc2+i);
SIMD<double> b0(pb0+i);
FNMAnonasm(a00, b0, c0);
FNMAnonasm(a10, b0, c1);
FNMAnonasm(a20, b0, c2);
SIMD<double> b1(pb1+i);
FNMAnonasm(a01, b1, c0);
FNMAnonasm(a11, b1, c1);
FNMAnonasm(a21, b1, c2);
SIMD<double> b2(pb2+i);
FNMAnonasm(a02, b2, c0);
FNMAnonasm(a12, b2, c1);
FNMAnonasm(a22, b2, c2);
SIMD<double> b3(pb3+i);
FNMAnonasm(a03, b3, c0);
FNMAnonasm(a13, b3, c1);
FNMAnonasm(a23, b3, c2);
c0.Store(pc0+i);
c1.Store(pc1+i);
c2.Store(pc2+i);
}
SIMD<mask64> mask(n%SW);
SIMD<double> c0(pc0+i, mask);
SIMD<double> c1(pc1+i, mask);
SIMD<double> c2(pc2+i, mask);
SIMD<double> b0(pb0+i, mask);
c0 -= a00 * b0;
c1 -= a10 * b0;
c2 -= a20 * b0;
SIMD<double> b1(pb1+i, mask);
c0 -= a01 * b1;
c1 -= a11 * b1;
c2 -= a21 * b1;
SIMD<double> b2(pb2+i, mask);
c0 -= a02 * b2;
c1 -= a12 * b2;
c2 -= a22 * b2;
SIMD<double> b3(pb3+i, mask);
c0 -= a03 * b3;
c1 -= a13 * b3;
c2 -= a23 * b3;
c0.Store(pc0+i, mask);
c1.Store(pc1+i, mask);
c2.Store(pc2+i, mask);
}
// C = A * B,  with short inner loop
template <size_t WA, OPERATION OP>
inline void MatKernelShortSum
(size_t ha, size_t wb, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
// C = A * B,  with short inner loop, unroll width B
template <size_t WA, OPERATION OP>
inline void MatKernelShortSum2
(size_t ha, size_t wb, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <> INLINE void MatKernelShortSum<0, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<0, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<0, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<0, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<0, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<0, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<0, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<0, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<1, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<1, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<1, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<1, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<1, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<1, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<1, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<1, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<2, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<2, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<2, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<2, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<2, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<2, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<2, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<2, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<3, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<3, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<3, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<3, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<3, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<3, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<3, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<3, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<4, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<4, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<4, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<4, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<4, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<4, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<4, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<4, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<5, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<5, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<5, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<5, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<5, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<5, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<5, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<5, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<6, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<6, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<6, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<6, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<6, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<6, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<6, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<6, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<7, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<7, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<7, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<7, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<7, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<7, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<7, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<7, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<8, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<8, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<8, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<8, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<8, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<8, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<8, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<8, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<9, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<9, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<9, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<9, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<9, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<9, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<9, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<9, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<10, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<10, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<10, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<10, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<10, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<10, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<10, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<10, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<11, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
FMAasm (b10,SIMD<double>(pa2[10]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<11, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
FMAasm(SIMD<double>(pa2[10]),b100, sum0);
FMAasm(SIMD<double>(pa2[10]),b101, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum += SIMD<double,1>(pa2[10]) * b10;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0 += SIMD<double>(pa2[10]) * b100;
sum1 += SIMD<double>(pa2[10]) * b101;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<11, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
FNMAasm (b10,SIMD<double>(pa2[10]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<11, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
FNMAasm(SIMD<double>(pa2[10]),b100, sum0);
FNMAasm(SIMD<double>(pa2[10]),b101, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum -= SIMD<double,1>(pa2[10]) * b10;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0 -= SIMD<double>(pa2[10]) * b100;
sum1 -= SIMD<double>(pa2[10]) * b101;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<11, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
FMAasm (b10,SIMD<double>(pa2[10]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<11, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
FMAasm(SIMD<double>(pa2[10]),b100, sum0);
FMAasm(SIMD<double>(pa2[10]),b101, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum += SIMD<double,1>(pa2[10]) * b10;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0 += SIMD<double>(pa2[10]) * b100;
sum1 += SIMD<double>(pa2[10]) * b101;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<11, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
FNMAasm (b10,SIMD<double>(pa2[10]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<11, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
FNMAasm(SIMD<double>(pa2[10]),b100, sum0);
FNMAasm(SIMD<double>(pa2[10]),b101, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum -= SIMD<double,1>(pa2[10]) * b10;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0 -= SIMD<double>(pa2[10]) * b100;
sum1 -= SIMD<double>(pa2[10]) * b101;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<12, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
FMAasm (b10,SIMD<double>(pa2[10]), sum);
FMAasm (b11,SIMD<double>(pa2[11]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<12, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
FMAasm(SIMD<double>(pa2[10]),b100, sum0);
FMAasm(SIMD<double>(pa2[10]),b101, sum1);
FMAasm(SIMD<double>(pa2[11]),b110, sum0);
FMAasm(SIMD<double>(pa2[11]),b111, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
SIMD<double,1> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum += SIMD<double,1>(pa2[10]) * b10;
sum += SIMD<double,1>(pa2[11]) * b11;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0 += SIMD<double>(pa2[10]) * b100;
sum1 += SIMD<double>(pa2[10]) * b101;
sum0 += SIMD<double>(pa2[11]) * b110;
sum1 += SIMD<double>(pa2[11]) * b111;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<12, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
FNMAasm (b10,SIMD<double>(pa2[10]), sum);
FNMAasm (b11,SIMD<double>(pa2[11]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<12, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
FNMAasm(SIMD<double>(pa2[10]),b100, sum0);
FNMAasm(SIMD<double>(pa2[10]),b101, sum1);
FNMAasm(SIMD<double>(pa2[11]),b110, sum0);
FNMAasm(SIMD<double>(pa2[11]),b111, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
SIMD<double,1> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum = 0.0;
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum -= SIMD<double,1>(pa2[10]) * b10;
sum -= SIMD<double,1>(pa2[11]) * b11;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0 = 0.0;
SIMD<double> sum1 = 0.0;
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0 -= SIMD<double>(pa2[10]) * b100;
sum1 -= SIMD<double>(pa2[10]) * b101;
sum0 -= SIMD<double>(pa2[11]) * b110;
sum1 -= SIMD<double>(pa2[11]) * b111;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum = 0.0;
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<12, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FMAasm (b0,SIMD<double>(pa2[0]), sum);
FMAasm (b1,SIMD<double>(pa2[1]), sum);
FMAasm (b2,SIMD<double>(pa2[2]), sum);
FMAasm (b3,SIMD<double>(pa2[3]), sum);
FMAasm (b4,SIMD<double>(pa2[4]), sum);
FMAasm (b5,SIMD<double>(pa2[5]), sum);
FMAasm (b6,SIMD<double>(pa2[6]), sum);
FMAasm (b7,SIMD<double>(pa2[7]), sum);
FMAasm (b8,SIMD<double>(pa2[8]), sum);
FMAasm (b9,SIMD<double>(pa2[9]), sum);
FMAasm (b10,SIMD<double>(pa2[10]), sum);
FMAasm (b11,SIMD<double>(pa2[11]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<12, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FMAasm(SIMD<double>(pa2[0]),b00, sum0);
FMAasm(SIMD<double>(pa2[0]),b01, sum1);
FMAasm(SIMD<double>(pa2[1]),b10, sum0);
FMAasm(SIMD<double>(pa2[1]),b11, sum1);
FMAasm(SIMD<double>(pa2[2]),b20, sum0);
FMAasm(SIMD<double>(pa2[2]),b21, sum1);
FMAasm(SIMD<double>(pa2[3]),b30, sum0);
FMAasm(SIMD<double>(pa2[3]),b31, sum1);
FMAasm(SIMD<double>(pa2[4]),b40, sum0);
FMAasm(SIMD<double>(pa2[4]),b41, sum1);
FMAasm(SIMD<double>(pa2[5]),b50, sum0);
FMAasm(SIMD<double>(pa2[5]),b51, sum1);
FMAasm(SIMD<double>(pa2[6]),b60, sum0);
FMAasm(SIMD<double>(pa2[6]),b61, sum1);
FMAasm(SIMD<double>(pa2[7]),b70, sum0);
FMAasm(SIMD<double>(pa2[7]),b71, sum1);
FMAasm(SIMD<double>(pa2[8]),b80, sum0);
FMAasm(SIMD<double>(pa2[8]),b81, sum1);
FMAasm(SIMD<double>(pa2[9]),b90, sum0);
FMAasm(SIMD<double>(pa2[9]),b91, sum1);
FMAasm(SIMD<double>(pa2[10]),b100, sum0);
FMAasm(SIMD<double>(pa2[10]),b101, sum1);
FMAasm(SIMD<double>(pa2[11]),b110, sum0);
FMAasm(SIMD<double>(pa2[11]),b111, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
SIMD<double,1> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum += SIMD<double,1>(pa2[0]) * b0;
sum += SIMD<double,1>(pa2[1]) * b1;
sum += SIMD<double,1>(pa2[2]) * b2;
sum += SIMD<double,1>(pa2[3]) * b3;
sum += SIMD<double,1>(pa2[4]) * b4;
sum += SIMD<double,1>(pa2[5]) * b5;
sum += SIMD<double,1>(pa2[6]) * b6;
sum += SIMD<double,1>(pa2[7]) * b7;
sum += SIMD<double,1>(pa2[8]) * b8;
sum += SIMD<double,1>(pa2[9]) * b9;
sum += SIMD<double,1>(pa2[10]) * b10;
sum += SIMD<double,1>(pa2[11]) * b11;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 += SIMD<double>(pa2[0]) * b00;
sum1 += SIMD<double>(pa2[0]) * b01;
sum0 += SIMD<double>(pa2[1]) * b10;
sum1 += SIMD<double>(pa2[1]) * b11;
sum0 += SIMD<double>(pa2[2]) * b20;
sum1 += SIMD<double>(pa2[2]) * b21;
sum0 += SIMD<double>(pa2[3]) * b30;
sum1 += SIMD<double>(pa2[3]) * b31;
sum0 += SIMD<double>(pa2[4]) * b40;
sum1 += SIMD<double>(pa2[4]) * b41;
sum0 += SIMD<double>(pa2[5]) * b50;
sum1 += SIMD<double>(pa2[5]) * b51;
sum0 += SIMD<double>(pa2[6]) * b60;
sum1 += SIMD<double>(pa2[6]) * b61;
sum0 += SIMD<double>(pa2[7]) * b70;
sum1 += SIMD<double>(pa2[7]) * b71;
sum0 += SIMD<double>(pa2[8]) * b80;
sum1 += SIMD<double>(pa2[8]) * b81;
sum0 += SIMD<double>(pa2[9]) * b90;
sum1 += SIMD<double>(pa2[9]) * b91;
sum0 += SIMD<double>(pa2[10]) * b100;
sum1 += SIMD<double>(pa2[10]) * b101;
sum0 += SIMD<double>(pa2[11]) * b110;
sum1 += SIMD<double>(pa2[11]) * b111;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum += SIMD<double>(pa2[0]) * b0;
sum += SIMD<double>(pa2[1]) * b1;
sum += SIMD<double>(pa2[2]) * b2;
sum += SIMD<double>(pa2[3]) * b3;
sum += SIMD<double>(pa2[4]) * b4;
sum += SIMD<double>(pa2[5]) * b5;
sum += SIMD<double>(pa2[6]) * b6;
sum += SIMD<double>(pa2[7]) * b7;
sum += SIMD<double>(pa2[8]) * b8;
sum += SIMD<double>(pa2[9]) * b9;
sum += SIMD<double>(pa2[10]) * b10;
sum += SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum<12, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
FNMAasm (b0,SIMD<double>(pa2[0]), sum);
FNMAasm (b1,SIMD<double>(pa2[1]), sum);
FNMAasm (b2,SIMD<double>(pa2[2]), sum);
FNMAasm (b3,SIMD<double>(pa2[3]), sum);
FNMAasm (b4,SIMD<double>(pa2[4]), sum);
FNMAasm (b5,SIMD<double>(pa2[5]), sum);
FNMAasm (b6,SIMD<double>(pa2[6]), sum);
FNMAasm (b7,SIMD<double>(pa2[7]), sum);
FNMAasm (b8,SIMD<double>(pa2[8]), sum);
FNMAasm (b9,SIMD<double>(pa2[9]), sum);
FNMAasm (b10,SIMD<double>(pa2[10]), sum);
FNMAasm (b11,SIMD<double>(pa2[11]), sum);
sum.Store(pc2);
} }
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
template <> INLINE void MatKernelShortSum2<12, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+2*SW <= wb; i += 2*SW, pb += 2*SW, pc += 2*SW)
{
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW);
FNMAasm(SIMD<double>(pa2[0]),b00, sum0);
FNMAasm(SIMD<double>(pa2[0]),b01, sum1);
FNMAasm(SIMD<double>(pa2[1]),b10, sum0);
FNMAasm(SIMD<double>(pa2[1]),b11, sum1);
FNMAasm(SIMD<double>(pa2[2]),b20, sum0);
FNMAasm(SIMD<double>(pa2[2]),b21, sum1);
FNMAasm(SIMD<double>(pa2[3]),b30, sum0);
FNMAasm(SIMD<double>(pa2[3]),b31, sum1);
FNMAasm(SIMD<double>(pa2[4]),b40, sum0);
FNMAasm(SIMD<double>(pa2[4]),b41, sum1);
FNMAasm(SIMD<double>(pa2[5]),b50, sum0);
FNMAasm(SIMD<double>(pa2[5]),b51, sum1);
FNMAasm(SIMD<double>(pa2[6]),b60, sum0);
FNMAasm(SIMD<double>(pa2[6]),b61, sum1);
FNMAasm(SIMD<double>(pa2[7]),b70, sum0);
FNMAasm(SIMD<double>(pa2[7]),b71, sum1);
FNMAasm(SIMD<double>(pa2[8]),b80, sum0);
FNMAasm(SIMD<double>(pa2[8]),b81, sum1);
FNMAasm(SIMD<double>(pa2[9]),b90, sum0);
FNMAasm(SIMD<double>(pa2[9]),b91, sum1);
FNMAasm(SIMD<double>(pa2[10]),b100, sum0);
FNMAasm(SIMD<double>(pa2[10]),b101, sum1);
FNMAasm(SIMD<double>(pa2[11]),b110, sum0);
FNMAasm(SIMD<double>(pa2[11]),b111, sum1);
sum0.Store(pc2);
sum1.Store(pc2+SW);
} }
size_t rest = wb % (2*SW); 
if (rest == 0) return; 
if (rest & 1) {  
double * pb2 = pb;
SIMD<double,1> b0(pb2); pb2 += db;
SIMD<double,1> b1(pb2); pb2 += db;
SIMD<double,1> b2(pb2); pb2 += db;
SIMD<double,1> b3(pb2); pb2 += db;
SIMD<double,1> b4(pb2); pb2 += db;
SIMD<double,1> b5(pb2); pb2 += db;
SIMD<double,1> b6(pb2); pb2 += db;
SIMD<double,1> b7(pb2); pb2 += db;
SIMD<double,1> b8(pb2); pb2 += db;
SIMD<double,1> b9(pb2); pb2 += db;
SIMD<double,1> b10(pb2); pb2 += db;
SIMD<double,1> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double,1> sum(pc2);
sum -= SIMD<double,1>(pa2[0]) * b0;
sum -= SIMD<double,1>(pa2[1]) * b1;
sum -= SIMD<double,1>(pa2[2]) * b2;
sum -= SIMD<double,1>(pa2[3]) * b3;
sum -= SIMD<double,1>(pa2[4]) * b4;
sum -= SIMD<double,1>(pa2[5]) * b5;
sum -= SIMD<double,1>(pa2[6]) * b6;
sum -= SIMD<double,1>(pa2[7]) * b7;
sum -= SIMD<double,1>(pa2[8]) * b8;
sum -= SIMD<double,1>(pa2[9]) * b9;
sum -= SIMD<double,1>(pa2[10]) * b10;
sum -= SIMD<double,1>(pa2[11]) * b11;
sum.Store(pc2);
}
pc += 1;
pb += 1;
}
return; 
if (rest >= SW) 
{
if (rest > SW)
{
SIMD<mask64> mask(rest-SW); 
double * pb2 = pb;
SIMD<double> b00(pb2);
SIMD<double> b01(pb2+SW,mask); pb2 += db;
SIMD<double> b10(pb2);
SIMD<double> b11(pb2+SW,mask); pb2 += db;
SIMD<double> b20(pb2);
SIMD<double> b21(pb2+SW,mask); pb2 += db;
SIMD<double> b30(pb2);
SIMD<double> b31(pb2+SW,mask); pb2 += db;
SIMD<double> b40(pb2);
SIMD<double> b41(pb2+SW,mask); pb2 += db;
SIMD<double> b50(pb2);
SIMD<double> b51(pb2+SW,mask); pb2 += db;
SIMD<double> b60(pb2);
SIMD<double> b61(pb2+SW,mask); pb2 += db;
SIMD<double> b70(pb2);
SIMD<double> b71(pb2+SW,mask); pb2 += db;
SIMD<double> b80(pb2);
SIMD<double> b81(pb2+SW,mask); pb2 += db;
SIMD<double> b90(pb2);
SIMD<double> b91(pb2+SW,mask); pb2 += db;
SIMD<double> b100(pb2);
SIMD<double> b101(pb2+SW,mask); pb2 += db;
SIMD<double> b110(pb2);
SIMD<double> b111(pb2+SW,mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum0(pc2);
SIMD<double> sum1(pc2+SW,mask);
sum0 -= SIMD<double>(pa2[0]) * b00;
sum1 -= SIMD<double>(pa2[0]) * b01;
sum0 -= SIMD<double>(pa2[1]) * b10;
sum1 -= SIMD<double>(pa2[1]) * b11;
sum0 -= SIMD<double>(pa2[2]) * b20;
sum1 -= SIMD<double>(pa2[2]) * b21;
sum0 -= SIMD<double>(pa2[3]) * b30;
sum1 -= SIMD<double>(pa2[3]) * b31;
sum0 -= SIMD<double>(pa2[4]) * b40;
sum1 -= SIMD<double>(pa2[4]) * b41;
sum0 -= SIMD<double>(pa2[5]) * b50;
sum1 -= SIMD<double>(pa2[5]) * b51;
sum0 -= SIMD<double>(pa2[6]) * b60;
sum1 -= SIMD<double>(pa2[6]) * b61;
sum0 -= SIMD<double>(pa2[7]) * b70;
sum1 -= SIMD<double>(pa2[7]) * b71;
sum0 -= SIMD<double>(pa2[8]) * b80;
sum1 -= SIMD<double>(pa2[8]) * b81;
sum0 -= SIMD<double>(pa2[9]) * b90;
sum1 -= SIMD<double>(pa2[9]) * b91;
sum0 -= SIMD<double>(pa2[10]) * b100;
sum1 -= SIMD<double>(pa2[10]) * b101;
sum0 -= SIMD<double>(pa2[11]) * b110;
sum1 -= SIMD<double>(pa2[11]) * b111;
sum0.Store(pc2);
sum1.Store(pc2+SW,mask);
}
return;
}
double * pb2 = pb;
SIMD<double> b0(pb2); pb2 += db;
SIMD<double> b1(pb2); pb2 += db;
SIMD<double> b2(pb2); pb2 += db;
SIMD<double> b3(pb2); pb2 += db;
SIMD<double> b4(pb2); pb2 += db;
SIMD<double> b5(pb2); pb2 += db;
SIMD<double> b6(pb2); pb2 += db;
SIMD<double> b7(pb2); pb2 += db;
SIMD<double> b8(pb2); pb2 += db;
SIMD<double> b9(pb2); pb2 += db;
SIMD<double> b10(pb2); pb2 += db;
SIMD<double> b11(pb2); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2);
}
return;
}
SIMD<mask64> mask(rest); 
double * pb2 = pb;
SIMD<double> b0(pb2, mask); pb2 += db;
SIMD<double> b1(pb2, mask); pb2 += db;
SIMD<double> b2(pb2, mask); pb2 += db;
SIMD<double> b3(pb2, mask); pb2 += db;
SIMD<double> b4(pb2, mask); pb2 += db;
SIMD<double> b5(pb2, mask); pb2 += db;
SIMD<double> b6(pb2, mask); pb2 += db;
SIMD<double> b7(pb2, mask); pb2 += db;
SIMD<double> b8(pb2, mask); pb2 += db;
SIMD<double> b9(pb2, mask); pb2 += db;
SIMD<double> b10(pb2, mask); pb2 += db;
SIMD<double> b11(pb2, mask); pb2 += db;
double * pa2 = pa;
double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pc2 += dc)
{
SIMD<double> sum(pc2, mask);
sum -= SIMD<double>(pa2[0]) * b0;
sum -= SIMD<double>(pa2[1]) * b1;
sum -= SIMD<double>(pa2[2]) * b2;
sum -= SIMD<double>(pa2[3]) * b3;
sum -= SIMD<double>(pa2[4]) * b4;
sum -= SIMD<double>(pa2[5]) * b5;
sum -= SIMD<double>(pa2[6]) * b6;
sum -= SIMD<double>(pa2[7]) * b7;
sum -= SIMD<double>(pa2[8]) * b8;
sum -= SIMD<double>(pa2[9]) * b9;
sum -= SIMD<double>(pa2[10]) * b10;
sum -= SIMD<double>(pa2[11]) * b11;
sum.Store(pc2, mask);
} }
// C = A^t * B,  with short inner loop
template <size_t WA, OPERATION OP>
inline void MatKernelAtB_SmallWA
(size_t ha, size_t wb, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
// C = A^t * B,  with short inner loop, unroll width B
template <size_t WA, OPERATION OP>
inline void MatKernelAtB_SmallWA2
(size_t ha, size_t wb, double * pa, size_t da, double * pb, size_t db, double * pc, size_t dc);
template <> INLINE void MatKernelAtB_SmallWA<0, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA<0, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA<0, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA<0, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA<1, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<1, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<1, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<1, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<2, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<2, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<2, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<2, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<3, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<3, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<3, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<3, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<4, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<4, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<4, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<4, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<5, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<5, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<5, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<5, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<6, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<6, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<6, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<6, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<7, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<7, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<7, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<7, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<8, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<8, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<8, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<8, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<9, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<9, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<9, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<9, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<10, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<10, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<10, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<10, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<11, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<11, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
SIMD<double> sum10(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
SIMD<double> sum10(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<11, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
SIMD<double> sum10(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
SIMD<double> sum10(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<11, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<12, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
sum11.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
sum11.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<12, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
SIMD<double> sum10(pc2); pc2 += dc;
SIMD<double> sum11(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
sum11.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
SIMD<double> sum10(pc2,mask); pc2 += dc;
SIMD<double> sum11(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
sum11.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<12, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
SIMD<double> sum8(pc2); pc2 += dc;
SIMD<double> sum9(pc2); pc2 += dc;
SIMD<double> sum10(pc2); pc2 += dc;
SIMD<double> sum11(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FNMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
sum11.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
SIMD<double> sum8(pc2,mask); pc2 += dc;
SIMD<double> sum9(pc2,mask); pc2 += dc;
SIMD<double> sum10(pc2,mask); pc2 += dc;
SIMD<double> sum11(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FNMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
sum11.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA<12, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
for (size_t i = 0; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FNMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
sum8.Store(pc2); pc2 += dc;
sum9.Store(pc2); pc2 += dc;
sum10.Store(pc2); pc2 += dc;
sum11.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
SIMD<double> sum8(0.0);
SIMD<double> sum9(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm(bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm(bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm(bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm(bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm(bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm(bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm(bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm(bjk,SIMD<double>(pa2[7]), sum7);
FNMAasm(bjk,SIMD<double>(pa2[8]), sum8);
FNMAasm(bjk,SIMD<double>(pa2[9]), sum9);
FNMAasm(bjk,SIMD<double>(pa2[10]), sum10);
FNMAasm(bjk,SIMD<double>(pa2[11]), sum11);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
sum8.Store(pc2, mask); pc2 += dc;
sum9.Store(pc2, mask); pc2 += dc;
sum10.Store(pc2, mask); pc2 += dc;
sum11.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<0, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
}
pc2 = pc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA2<0, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
}
pc2 = pc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA2<0, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
}
pc2 = pc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA2<0, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
}
pc2 = pc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
}
pc2 = pc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
}
pc2 = pc;
}
template <> INLINE void MatKernelAtB_SmallWA2<1, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<1, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<1, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<1, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<2, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<2, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<2, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<2, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<3, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<3, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<3, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<3, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<4, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<4, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<4, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<4, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<5, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<5, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<5, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<5, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<6, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<6, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<6, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<6, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<7, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
SIMD<double> sum60(0.0);
SIMD<double> sum61(0.0);
SIMD<double> sum62(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<7, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
SIMD<double> sum60(pc2);
SIMD<double> sum61(pc2+SW);
SIMD<double> sum62(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<7, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
SIMD<double> sum60(pc2);
SIMD<double> sum61(pc2+SW);
SIMD<double> sum62(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FNMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FNMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FNMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<7, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
SIMD<double> sum60(0.0);
SIMD<double> sum61(0.0);
SIMD<double> sum62(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FNMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FNMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FNMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<8, SET>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
SIMD<double> sum60(0.0);
SIMD<double> sum61(0.0);
SIMD<double> sum62(0.0);
SIMD<double> sum70(0.0);
SIMD<double> sum71(0.0);
SIMD<double> sum72(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
FMAasm(bjk0,SIMD<double>(pa2[7]), sum70);
FMAasm(bjk1,SIMD<double>(pa2[7]), sum71);
FMAasm(bjk2,SIMD<double>(pa2[7]), sum72);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
sum70.Store(pc2);
sum71.Store(pc2+SW);
sum72.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<8, ADD>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
SIMD<double> sum60(pc2);
SIMD<double> sum61(pc2+SW);
SIMD<double> sum62(pc2+2*SW); pc2 += dc;
SIMD<double> sum70(pc2);
SIMD<double> sum71(pc2+SW);
SIMD<double> sum72(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
FMAasm(bjk0,SIMD<double>(pa2[7]), sum70);
FMAasm(bjk1,SIMD<double>(pa2[7]), sum71);
FMAasm(bjk2,SIMD<double>(pa2[7]), sum72);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
sum70.Store(pc2);
sum71.Store(pc2+SW);
sum72.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<8, SUB>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(pc2);
SIMD<double> sum01(pc2+SW);
SIMD<double> sum02(pc2+2*SW); pc2 += dc;
SIMD<double> sum10(pc2);
SIMD<double> sum11(pc2+SW);
SIMD<double> sum12(pc2+2*SW); pc2 += dc;
SIMD<double> sum20(pc2);
SIMD<double> sum21(pc2+SW);
SIMD<double> sum22(pc2+2*SW); pc2 += dc;
SIMD<double> sum30(pc2);
SIMD<double> sum31(pc2+SW);
SIMD<double> sum32(pc2+2*SW); pc2 += dc;
SIMD<double> sum40(pc2);
SIMD<double> sum41(pc2+SW);
SIMD<double> sum42(pc2+2*SW); pc2 += dc;
SIMD<double> sum50(pc2);
SIMD<double> sum51(pc2+SW);
SIMD<double> sum52(pc2+2*SW); pc2 += dc;
SIMD<double> sum60(pc2);
SIMD<double> sum61(pc2+SW);
SIMD<double> sum62(pc2+2*SW); pc2 += dc;
SIMD<double> sum70(pc2);
SIMD<double> sum71(pc2+SW);
SIMD<double> sum72(pc2+2*SW); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FNMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FNMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FNMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
FNMAasm(bjk0,SIMD<double>(pa2[7]), sum70);
FNMAasm(bjk1,SIMD<double>(pa2[7]), sum71);
FNMAasm(bjk2,SIMD<double>(pa2[7]), sum72);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
sum70.Store(pc2);
sum71.Store(pc2+SW);
sum72.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2); pc2 += dc;
SIMD<double> sum1(pc2); pc2 += dc;
SIMD<double> sum2(pc2); pc2 += dc;
SIMD<double> sum3(pc2); pc2 += dc;
SIMD<double> sum4(pc2); pc2 += dc;
SIMD<double> sum5(pc2); pc2 += dc;
SIMD<double> sum6(pc2); pc2 += dc;
SIMD<double> sum7(pc2); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(pc2,mask); pc2 += dc;
SIMD<double> sum1(pc2,mask); pc2 += dc;
SIMD<double> sum2(pc2,mask); pc2 += dc;
SIMD<double> sum3(pc2,mask); pc2 += dc;
SIMD<double> sum4(pc2,mask); pc2 += dc;
SIMD<double> sum5(pc2,mask); pc2 += dc;
SIMD<double> sum6(pc2,mask); pc2 += dc;
SIMD<double> sum7(pc2,mask); pc2 += dc;
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
template <> INLINE void MatKernelAtB_SmallWA2<8, SETNEG>
    (size_t ha, size_t wb,
     double * pa, size_t da,
     double * pb, size_t db,
     double * pc, size_t dc)
{
constexpr int SW = SIMD<double>::Size();
size_t i = 0;
for ( ; i+3*SW <= wb; i += 3*SW, pb += 3*SW, pc += 3*SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum00(0.0);
SIMD<double> sum01(0.0);
SIMD<double> sum02(0.0);
SIMD<double> sum10(0.0);
SIMD<double> sum11(0.0);
SIMD<double> sum12(0.0);
SIMD<double> sum20(0.0);
SIMD<double> sum21(0.0);
SIMD<double> sum22(0.0);
SIMD<double> sum30(0.0);
SIMD<double> sum31(0.0);
SIMD<double> sum32(0.0);
SIMD<double> sum40(0.0);
SIMD<double> sum41(0.0);
SIMD<double> sum42(0.0);
SIMD<double> sum50(0.0);
SIMD<double> sum51(0.0);
SIMD<double> sum52(0.0);
SIMD<double> sum60(0.0);
SIMD<double> sum61(0.0);
SIMD<double> sum62(0.0);
SIMD<double> sum70(0.0);
SIMD<double> sum71(0.0);
SIMD<double> sum72(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk0(pb2);
SIMD<double> bjk1(pb2+SW);
SIMD<double> bjk2(pb2+2*SW);
FNMAasm(bjk0,SIMD<double>(pa2[0]), sum00);
FNMAasm(bjk1,SIMD<double>(pa2[0]), sum01);
FNMAasm(bjk2,SIMD<double>(pa2[0]), sum02);
FNMAasm(bjk0,SIMD<double>(pa2[1]), sum10);
FNMAasm(bjk1,SIMD<double>(pa2[1]), sum11);
FNMAasm(bjk2,SIMD<double>(pa2[1]), sum12);
FNMAasm(bjk0,SIMD<double>(pa2[2]), sum20);
FNMAasm(bjk1,SIMD<double>(pa2[2]), sum21);
FNMAasm(bjk2,SIMD<double>(pa2[2]), sum22);
FNMAasm(bjk0,SIMD<double>(pa2[3]), sum30);
FNMAasm(bjk1,SIMD<double>(pa2[3]), sum31);
FNMAasm(bjk2,SIMD<double>(pa2[3]), sum32);
FNMAasm(bjk0,SIMD<double>(pa2[4]), sum40);
FNMAasm(bjk1,SIMD<double>(pa2[4]), sum41);
FNMAasm(bjk2,SIMD<double>(pa2[4]), sum42);
FNMAasm(bjk0,SIMD<double>(pa2[5]), sum50);
FNMAasm(bjk1,SIMD<double>(pa2[5]), sum51);
FNMAasm(bjk2,SIMD<double>(pa2[5]), sum52);
FNMAasm(bjk0,SIMD<double>(pa2[6]), sum60);
FNMAasm(bjk1,SIMD<double>(pa2[6]), sum61);
FNMAasm(bjk2,SIMD<double>(pa2[6]), sum62);
FNMAasm(bjk0,SIMD<double>(pa2[7]), sum70);
FNMAasm(bjk1,SIMD<double>(pa2[7]), sum71);
FNMAasm(bjk2,SIMD<double>(pa2[7]), sum72);
}
pc2 = pc;
sum00.Store(pc2);
sum01.Store(pc2+SW);
sum02.Store(pc2+2*SW); pc2 += dc;
sum10.Store(pc2);
sum11.Store(pc2+SW);
sum12.Store(pc2+2*SW); pc2 += dc;
sum20.Store(pc2);
sum21.Store(pc2+SW);
sum22.Store(pc2+2*SW); pc2 += dc;
sum30.Store(pc2);
sum31.Store(pc2+SW);
sum32.Store(pc2+2*SW); pc2 += dc;
sum40.Store(pc2);
sum41.Store(pc2+SW);
sum42.Store(pc2+2*SW); pc2 += dc;
sum50.Store(pc2);
sum51.Store(pc2+SW);
sum52.Store(pc2+2*SW); pc2 += dc;
sum60.Store(pc2);
sum61.Store(pc2+SW);
sum62.Store(pc2+2*SW); pc2 += dc;
sum70.Store(pc2);
sum71.Store(pc2+SW);
sum72.Store(pc2+2*SW); pc2 += dc;
}
for ( ; i+SW <= wb; i += SW, pb += SW, pc += SW)
{
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2); pc2 += dc;
sum1.Store(pc2); pc2 += dc;
sum2.Store(pc2); pc2 += dc;
sum3.Store(pc2); pc2 += dc;
sum4.Store(pc2); pc2 += dc;
sum5.Store(pc2); pc2 += dc;
sum6.Store(pc2); pc2 += dc;
sum7.Store(pc2); pc2 += dc;
}
size_t rest = wb % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest); 
double * pb2 = pb;
double * pa2 = pa;
[[maybe_unused]] double * pc2 = pc;
__assume(ha>0);
SIMD<double> sum0(0.0);
SIMD<double> sum1(0.0);
SIMD<double> sum2(0.0);
SIMD<double> sum3(0.0);
SIMD<double> sum4(0.0);
SIMD<double> sum5(0.0);
SIMD<double> sum6(0.0);
SIMD<double> sum7(0.0);
#pragma unroll 1
for (size_t j = 0; j < ha; j++, pa2 += da, pb2 += db)
{
SIMD<double> bjk(pb2, mask);
FNMAasm (bjk,SIMD<double>(pa2[0]), sum0);
FNMAasm (bjk,SIMD<double>(pa2[1]), sum1);
FNMAasm (bjk,SIMD<double>(pa2[2]), sum2);
FNMAasm (bjk,SIMD<double>(pa2[3]), sum3);
FNMAasm (bjk,SIMD<double>(pa2[4]), sum4);
FNMAasm (bjk,SIMD<double>(pa2[5]), sum5);
FNMAasm (bjk,SIMD<double>(pa2[6]), sum6);
FNMAasm (bjk,SIMD<double>(pa2[7]), sum7);
}
pc2 = pc;
sum0.Store(pc2, mask); pc2 += dc;
sum1.Store(pc2, mask); pc2 += dc;
sum2.Store(pc2, mask); pc2 += dc;
sum3.Store(pc2, mask); pc2 += dc;
sum4.Store(pc2, mask); pc2 += dc;
sum5.Store(pc2, mask); pc2 += dc;
sum6.Store(pc2, mask); pc2 += dc;
sum7.Store(pc2, mask); pc2 += dc;
}
// y = A * x,  with fix width
template <size_t WA, OPERATION OP>
inline void KernelMatVec
(size_t ha, double * pa, size_t da, double * x, double * y);
template <> INLINE void KernelMatVec<0, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<0, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<1, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<1, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<2, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<2, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<3, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<3, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<4, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<4, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<5, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<5, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<6, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<6, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<7, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<7, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<8, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<8, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<9, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<9, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<10, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<10, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<11, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<11, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<12, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<12, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<13, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<13, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<14, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<14, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<15, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<15, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<16, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<16, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<17, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<17, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<18, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<18, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<19, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<19, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<20, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<20, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<21, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<21, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<22, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<22, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<23, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<23, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
y[i] += HSum(sum);
} }
template <> INLINE void KernelMatVec<24, SET>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
SIMD<double,1> x23(x+23);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
sum2 += SIMD<double,1>(pa+2*da+23) * x23;
sum3 += SIMD<double,1>(pa+3*da+23) * x23;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
sum += SIMD<double,1>(pa+23) * x23;
y[i] = HSum(sum);
} }
template <> INLINE void KernelMatVec<24, ADD>
(size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
SIMD<double,1> x23(x+23);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
sum2 += SIMD<double,1>(pa+2*da+23) * x23;
sum3 += SIMD<double,1>(pa+3*da+23) * x23;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum += SIMD<double,4> (y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum += SIMD<double,2> (y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
sum += SIMD<double,1>(pa+23) * x23;
y[i] += HSum(sum);
} }
// y += s * A * x,  with fix width
template <size_t WA>
inline void KernelAddMatVec
(double s, size_t ha, double * pa, size_t da, double * x, double * y);
template <> INLINE void KernelAddMatVec<0>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<1>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<2>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<3>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<4>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<5>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<6>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<7>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<8>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<9>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<10>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<11>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<12>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<13>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<14>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<15>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<16>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<17>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<18>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<19>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<20>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<21>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<22>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<23>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
y[i] += s*HSum(sum);
} }
template <> INLINE void KernelAddMatVec<24>
(double s, size_t ha, double * pa, size_t da, double * x, double * y) {
SIMD<double,1> x0(x+0);
SIMD<double,1> x1(x+1);
SIMD<double,1> x2(x+2);
SIMD<double,1> x3(x+3);
SIMD<double,1> x4(x+4);
SIMD<double,1> x5(x+5);
SIMD<double,1> x6(x+6);
SIMD<double,1> x7(x+7);
SIMD<double,1> x8(x+8);
SIMD<double,1> x9(x+9);
SIMD<double,1> x10(x+10);
SIMD<double,1> x11(x+11);
SIMD<double,1> x12(x+12);
SIMD<double,1> x13(x+13);
SIMD<double,1> x14(x+14);
SIMD<double,1> x15(x+15);
SIMD<double,1> x16(x+16);
SIMD<double,1> x17(x+17);
SIMD<double,1> x18(x+18);
SIMD<double,1> x19(x+19);
SIMD<double,1> x20(x+20);
SIMD<double,1> x21(x+21);
SIMD<double,1> x22(x+22);
SIMD<double,1> x23(x+23);
size_t i = 0;
for ( ; i+4 <= ha; i+=4, pa += 4*da) {
SIMD<double,1> sum0(0.0), sum1(0.0), sum2(0.0), sum3(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum2 += SIMD<double,1>(pa+2*da+0) * x0;
sum3 += SIMD<double,1>(pa+3*da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum2 += SIMD<double,1>(pa+2*da+1) * x1;
sum3 += SIMD<double,1>(pa+3*da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum2 += SIMD<double,1>(pa+2*da+2) * x2;
sum3 += SIMD<double,1>(pa+3*da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum2 += SIMD<double,1>(pa+2*da+3) * x3;
sum3 += SIMD<double,1>(pa+3*da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum2 += SIMD<double,1>(pa+2*da+4) * x4;
sum3 += SIMD<double,1>(pa+3*da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum2 += SIMD<double,1>(pa+2*da+5) * x5;
sum3 += SIMD<double,1>(pa+3*da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum2 += SIMD<double,1>(pa+2*da+6) * x6;
sum3 += SIMD<double,1>(pa+3*da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum2 += SIMD<double,1>(pa+2*da+7) * x7;
sum3 += SIMD<double,1>(pa+3*da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum2 += SIMD<double,1>(pa+2*da+8) * x8;
sum3 += SIMD<double,1>(pa+3*da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum2 += SIMD<double,1>(pa+2*da+9) * x9;
sum3 += SIMD<double,1>(pa+3*da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum2 += SIMD<double,1>(pa+2*da+10) * x10;
sum3 += SIMD<double,1>(pa+3*da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum2 += SIMD<double,1>(pa+2*da+11) * x11;
sum3 += SIMD<double,1>(pa+3*da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum2 += SIMD<double,1>(pa+2*da+12) * x12;
sum3 += SIMD<double,1>(pa+3*da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum2 += SIMD<double,1>(pa+2*da+13) * x13;
sum3 += SIMD<double,1>(pa+3*da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum2 += SIMD<double,1>(pa+2*da+14) * x14;
sum3 += SIMD<double,1>(pa+3*da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum2 += SIMD<double,1>(pa+2*da+15) * x15;
sum3 += SIMD<double,1>(pa+3*da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum2 += SIMD<double,1>(pa+2*da+16) * x16;
sum3 += SIMD<double,1>(pa+3*da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum2 += SIMD<double,1>(pa+2*da+17) * x17;
sum3 += SIMD<double,1>(pa+3*da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum2 += SIMD<double,1>(pa+2*da+18) * x18;
sum3 += SIMD<double,1>(pa+3*da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum2 += SIMD<double,1>(pa+2*da+19) * x19;
sum3 += SIMD<double,1>(pa+3*da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum2 += SIMD<double,1>(pa+2*da+20) * x20;
sum3 += SIMD<double,1>(pa+3*da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum2 += SIMD<double,1>(pa+2*da+21) * x21;
sum3 += SIMD<double,1>(pa+3*da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum2 += SIMD<double,1>(pa+2*da+22) * x22;
sum3 += SIMD<double,1>(pa+3*da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
sum2 += SIMD<double,1>(pa+2*da+23) * x23;
sum3 += SIMD<double,1>(pa+3*da+23) * x23;
SIMD<double,4> vsum = HSum(sum0,sum1,sum2,sum3);
vsum *= SIMD<double,4>(s);
vsum += SIMD<double,4>(y+i);
vsum.Store(y+i);
}
if (ha & 2) {
SIMD<double,1> sum0(0.0), sum1(0.0);
sum0 += SIMD<double,1>(pa+0) * x0;
sum1 += SIMD<double,1>(pa+da+0) * x0;
sum0 += SIMD<double,1>(pa+1) * x1;
sum1 += SIMD<double,1>(pa+da+1) * x1;
sum0 += SIMD<double,1>(pa+2) * x2;
sum1 += SIMD<double,1>(pa+da+2) * x2;
sum0 += SIMD<double,1>(pa+3) * x3;
sum1 += SIMD<double,1>(pa+da+3) * x3;
sum0 += SIMD<double,1>(pa+4) * x4;
sum1 += SIMD<double,1>(pa+da+4) * x4;
sum0 += SIMD<double,1>(pa+5) * x5;
sum1 += SIMD<double,1>(pa+da+5) * x5;
sum0 += SIMD<double,1>(pa+6) * x6;
sum1 += SIMD<double,1>(pa+da+6) * x6;
sum0 += SIMD<double,1>(pa+7) * x7;
sum1 += SIMD<double,1>(pa+da+7) * x7;
sum0 += SIMD<double,1>(pa+8) * x8;
sum1 += SIMD<double,1>(pa+da+8) * x8;
sum0 += SIMD<double,1>(pa+9) * x9;
sum1 += SIMD<double,1>(pa+da+9) * x9;
sum0 += SIMD<double,1>(pa+10) * x10;
sum1 += SIMD<double,1>(pa+da+10) * x10;
sum0 += SIMD<double,1>(pa+11) * x11;
sum1 += SIMD<double,1>(pa+da+11) * x11;
sum0 += SIMD<double,1>(pa+12) * x12;
sum1 += SIMD<double,1>(pa+da+12) * x12;
sum0 += SIMD<double,1>(pa+13) * x13;
sum1 += SIMD<double,1>(pa+da+13) * x13;
sum0 += SIMD<double,1>(pa+14) * x14;
sum1 += SIMD<double,1>(pa+da+14) * x14;
sum0 += SIMD<double,1>(pa+15) * x15;
sum1 += SIMD<double,1>(pa+da+15) * x15;
sum0 += SIMD<double,1>(pa+16) * x16;
sum1 += SIMD<double,1>(pa+da+16) * x16;
sum0 += SIMD<double,1>(pa+17) * x17;
sum1 += SIMD<double,1>(pa+da+17) * x17;
sum0 += SIMD<double,1>(pa+18) * x18;
sum1 += SIMD<double,1>(pa+da+18) * x18;
sum0 += SIMD<double,1>(pa+19) * x19;
sum1 += SIMD<double,1>(pa+da+19) * x19;
sum0 += SIMD<double,1>(pa+20) * x20;
sum1 += SIMD<double,1>(pa+da+20) * x20;
sum0 += SIMD<double,1>(pa+21) * x21;
sum1 += SIMD<double,1>(pa+da+21) * x21;
sum0 += SIMD<double,1>(pa+22) * x22;
sum1 += SIMD<double,1>(pa+da+22) * x22;
sum0 += SIMD<double,1>(pa+23) * x23;
sum1 += SIMD<double,1>(pa+da+23) * x23;
SIMD<double,2> vsum = HSum(sum0,sum1);
vsum *= SIMD<double,2>(s);
vsum += SIMD<double,2>(y+i);
vsum.Store(y+i);
i += 2; pa += 2*da;
}
if (ha & 1) {
SIMD<double,1> sum(0.0);
sum += SIMD<double,1>(pa+0) * x0;
sum += SIMD<double,1>(pa+1) * x1;
sum += SIMD<double,1>(pa+2) * x2;
sum += SIMD<double,1>(pa+3) * x3;
sum += SIMD<double,1>(pa+4) * x4;
sum += SIMD<double,1>(pa+5) * x5;
sum += SIMD<double,1>(pa+6) * x6;
sum += SIMD<double,1>(pa+7) * x7;
sum += SIMD<double,1>(pa+8) * x8;
sum += SIMD<double,1>(pa+9) * x9;
sum += SIMD<double,1>(pa+10) * x10;
sum += SIMD<double,1>(pa+11) * x11;
sum += SIMD<double,1>(pa+12) * x12;
sum += SIMD<double,1>(pa+13) * x13;
sum += SIMD<double,1>(pa+14) * x14;
sum += SIMD<double,1>(pa+15) * x15;
sum += SIMD<double,1>(pa+16) * x16;
sum += SIMD<double,1>(pa+17) * x17;
sum += SIMD<double,1>(pa+18) * x18;
sum += SIMD<double,1>(pa+19) * x19;
sum += SIMD<double,1>(pa+20) * x20;
sum += SIMD<double,1>(pa+21) * x21;
sum += SIMD<double,1>(pa+22) * x22;
sum += SIMD<double,1>(pa+23) * x23;
y[i] += s*HSum(sum);
} }
// y += s * A^t * x(ind),  with fix width
template <size_t WA>
inline void KernelAddMatTransVecI
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind);
template <>
inline void KernelAddMatTransVecI<0>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
size_t i;
for (i = 0; i+4<= ha; i+=4) {
SIMD<double> sx0(x[ind[i+0]]);
pa += da;
SIMD<double> sx1(x[ind[i+1]]);
pa += da;
SIMD<double> sx2(x[ind[i+2]]);
pa += da;
SIMD<double> sx3(x[ind[i+3]]);
pa += da;
}
switch (ha-i) {
case 3: {
SIMD<double> sx(x[ind[i]]);
pa += da; i++; }
case 2: {
SIMD<double> sx(x[ind[i]]);
pa += da; i++; }
case 1: {
SIMD<double> sx(x[ind[i]]);
pa += da; i++; }
default: ;}; 
; }
template <>
inline void KernelAddMatTransVecI<1>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy01(0);
SIMD<double> sy02(0);
SIMD<double> sy03(0);
size_t i;
for (i = 0; i+4<= ha; i+=4) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
pa += da;
SIMD<double> sx1(x[ind[i+1]]);
sy01 += SIMD<double>(pa+0) * sx1;
pa += da;
SIMD<double> sx2(x[ind[i+2]]);
sy02 += SIMD<double>(pa+0) * sx2;
pa += da;
SIMD<double> sx3(x[ind[i+3]]);
sy03 += SIMD<double>(pa+0) * sx3;
pa += da;
}
switch (ha-i) {
case 3: {
SIMD<double> sx(x[ind[i]]);
sy03 += SIMD<double>(pa+0) * sx;
pa += da; i++; }
case 2: {
SIMD<double> sx(x[ind[i]]);
sy02 += SIMD<double>(pa+0) * sx;
pa += da; i++; }
case 1: {
SIMD<double> sx(x[ind[i]]);
sy01 += SIMD<double>(pa+0) * sx;
pa += da; i++; }
default: ;}; 
sy00 += sy01+sy02+sy03;
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
; }
template <>
inline void KernelAddMatTransVecI<2>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy01(0);
SIMD<double> sy11(0);
size_t i;
for (i = 0; i+2<= ha; i+=2) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
pa += da;
SIMD<double> sx1(x[ind[i+1]]);
sy01 += SIMD<double>(pa+0) * sx1;
sy11 += SIMD<double>(pa+1) * sx1;
pa += da;
}
switch (ha-i) {
case 1: {
SIMD<double> sx(x[ind[i]]);
sy01 += SIMD<double>(pa+0) * sx;
sy11 += SIMD<double>(pa+1) * sx;
pa += da; i++; }
default: ;}; 
sy00 += sy01;
sy10 += sy11;
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
; }
template <>
inline void KernelAddMatTransVecI<3>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy01(0);
SIMD<double> sy11(0);
SIMD<double> sy21(0);
size_t i;
for (i = 0; i+2<= ha; i+=2) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
pa += da;
SIMD<double> sx1(x[ind[i+1]]);
sy01 += SIMD<double>(pa+0) * sx1;
sy11 += SIMD<double>(pa+1) * sx1;
sy21 += SIMD<double>(pa+2) * sx1;
pa += da;
}
switch (ha-i) {
case 1: {
SIMD<double> sx(x[ind[i]]);
sy01 += SIMD<double>(pa+0) * sx;
sy11 += SIMD<double>(pa+1) * sx;
sy21 += SIMD<double>(pa+2) * sx;
pa += da; i++; }
default: ;}; 
sy00 += sy01;
sy10 += sy11;
sy20 += sy21;
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
; }
template <>
inline void KernelAddMatTransVecI<4>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy01(0);
SIMD<double> sy11(0);
SIMD<double> sy21(0);
SIMD<double> sy31(0);
size_t i;
for (i = 0; i+2<= ha; i+=2) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
pa += da;
SIMD<double> sx1(x[ind[i+1]]);
sy01 += SIMD<double>(pa+0) * sx1;
sy11 += SIMD<double>(pa+1) * sx1;
sy21 += SIMD<double>(pa+2) * sx1;
sy31 += SIMD<double>(pa+3) * sx1;
pa += da;
}
switch (ha-i) {
case 1: {
SIMD<double> sx(x[ind[i]]);
sy01 += SIMD<double>(pa+0) * sx;
sy11 += SIMD<double>(pa+1) * sx;
sy21 += SIMD<double>(pa+2) * sx;
sy31 += SIMD<double>(pa+3) * sx;
pa += da; i++; }
default: ;}; 
sy00 += sy01;
sy10 += sy11;
sy20 += sy21;
sy30 += sy31;
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
; }
template <>
inline void KernelAddMatTransVecI<5>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
; }
template <>
inline void KernelAddMatTransVecI<6>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
; }
template <>
inline void KernelAddMatTransVecI<7>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
; }
template <>
inline void KernelAddMatTransVecI<8>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
; }
template <>
inline void KernelAddMatTransVecI<9>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
; }
template <>
inline void KernelAddMatTransVecI<10>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
; }
template <>
inline void KernelAddMatTransVecI<11>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
; }
template <>
inline void KernelAddMatTransVecI<12>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
; }
template <>
inline void KernelAddMatTransVecI<13>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
; }
template <>
inline void KernelAddMatTransVecI<14>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
; }
template <>
inline void KernelAddMatTransVecI<15>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
; }
template <>
inline void KernelAddMatTransVecI<16>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
; }
template <>
inline void KernelAddMatTransVecI<17>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
; }
template <>
inline void KernelAddMatTransVecI<18>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
; }
template <>
inline void KernelAddMatTransVecI<19>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
; }
template <>
inline void KernelAddMatTransVecI<20>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
SIMD<double> sy190(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
sy190 += SIMD<double>(pa+19) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
(s * sy190 + SIMD<double>(y+19)).Store(y+19);
; }
template <>
inline void KernelAddMatTransVecI<21>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
SIMD<double> sy190(0);
SIMD<double> sy200(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
sy190 += SIMD<double>(pa+19) * sx0;
sy200 += SIMD<double>(pa+20) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
(s * sy190 + SIMD<double>(y+19)).Store(y+19);
(s * sy200 + SIMD<double>(y+20)).Store(y+20);
; }
template <>
inline void KernelAddMatTransVecI<22>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
SIMD<double> sy190(0);
SIMD<double> sy200(0);
SIMD<double> sy210(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
sy190 += SIMD<double>(pa+19) * sx0;
sy200 += SIMD<double>(pa+20) * sx0;
sy210 += SIMD<double>(pa+21) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
(s * sy190 + SIMD<double>(y+19)).Store(y+19);
(s * sy200 + SIMD<double>(y+20)).Store(y+20);
(s * sy210 + SIMD<double>(y+21)).Store(y+21);
; }
template <>
inline void KernelAddMatTransVecI<23>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
SIMD<double> sy190(0);
SIMD<double> sy200(0);
SIMD<double> sy210(0);
SIMD<double> sy220(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
sy190 += SIMD<double>(pa+19) * sx0;
sy200 += SIMD<double>(pa+20) * sx0;
sy210 += SIMD<double>(pa+21) * sx0;
sy220 += SIMD<double>(pa+22) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
(s * sy190 + SIMD<double>(y+19)).Store(y+19);
(s * sy200 + SIMD<double>(y+20)).Store(y+20);
(s * sy210 + SIMD<double>(y+21)).Store(y+21);
(s * sy220 + SIMD<double>(y+22)).Store(y+22);
; }
template <>
inline void KernelAddMatTransVecI<24>
(double s, size_t ha, double * pa, size_t da, double * x, double * y, int * ind) {
SIMD<double> sy00(0);
SIMD<double> sy10(0);
SIMD<double> sy20(0);
SIMD<double> sy30(0);
SIMD<double> sy40(0);
SIMD<double> sy50(0);
SIMD<double> sy60(0);
SIMD<double> sy70(0);
SIMD<double> sy80(0);
SIMD<double> sy90(0);
SIMD<double> sy100(0);
SIMD<double> sy110(0);
SIMD<double> sy120(0);
SIMD<double> sy130(0);
SIMD<double> sy140(0);
SIMD<double> sy150(0);
SIMD<double> sy160(0);
SIMD<double> sy170(0);
SIMD<double> sy180(0);
SIMD<double> sy190(0);
SIMD<double> sy200(0);
SIMD<double> sy210(0);
SIMD<double> sy220(0);
SIMD<double> sy230(0);
size_t i;
for (i = 0; i+1<= ha; i+=1) {
SIMD<double> sx0(x[ind[i+0]]);
sy00 += SIMD<double>(pa+0) * sx0;
sy10 += SIMD<double>(pa+1) * sx0;
sy20 += SIMD<double>(pa+2) * sx0;
sy30 += SIMD<double>(pa+3) * sx0;
sy40 += SIMD<double>(pa+4) * sx0;
sy50 += SIMD<double>(pa+5) * sx0;
sy60 += SIMD<double>(pa+6) * sx0;
sy70 += SIMD<double>(pa+7) * sx0;
sy80 += SIMD<double>(pa+8) * sx0;
sy90 += SIMD<double>(pa+9) * sx0;
sy100 += SIMD<double>(pa+10) * sx0;
sy110 += SIMD<double>(pa+11) * sx0;
sy120 += SIMD<double>(pa+12) * sx0;
sy130 += SIMD<double>(pa+13) * sx0;
sy140 += SIMD<double>(pa+14) * sx0;
sy150 += SIMD<double>(pa+15) * sx0;
sy160 += SIMD<double>(pa+16) * sx0;
sy170 += SIMD<double>(pa+17) * sx0;
sy180 += SIMD<double>(pa+18) * sx0;
sy190 += SIMD<double>(pa+19) * sx0;
sy200 += SIMD<double>(pa+20) * sx0;
sy210 += SIMD<double>(pa+21) * sx0;
sy220 += SIMD<double>(pa+22) * sx0;
sy230 += SIMD<double>(pa+23) * sx0;
pa += da;
}
switch (ha-i) {
default: ;}; 
(s * sy00 + SIMD<double>(y+0)).Store(y+0);
(s * sy10 + SIMD<double>(y+1)).Store(y+1);
(s * sy20 + SIMD<double>(y+2)).Store(y+2);
(s * sy30 + SIMD<double>(y+3)).Store(y+3);
(s * sy40 + SIMD<double>(y+4)).Store(y+4);
(s * sy50 + SIMD<double>(y+5)).Store(y+5);
(s * sy60 + SIMD<double>(y+6)).Store(y+6);
(s * sy70 + SIMD<double>(y+7)).Store(y+7);
(s * sy80 + SIMD<double>(y+8)).Store(y+8);
(s * sy90 + SIMD<double>(y+9)).Store(y+9);
(s * sy100 + SIMD<double>(y+10)).Store(y+10);
(s * sy110 + SIMD<double>(y+11)).Store(y+11);
(s * sy120 + SIMD<double>(y+12)).Store(y+12);
(s * sy130 + SIMD<double>(y+13)).Store(y+13);
(s * sy140 + SIMD<double>(y+14)).Store(y+14);
(s * sy150 + SIMD<double>(y+15)).Store(y+15);
(s * sy160 + SIMD<double>(y+16)).Store(y+16);
(s * sy170 + SIMD<double>(y+17)).Store(y+17);
(s * sy180 + SIMD<double>(y+18)).Store(y+18);
(s * sy190 + SIMD<double>(y+19)).Store(y+19);
(s * sy200 + SIMD<double>(y+20)).Store(y+20);
(s * sy210 + SIMD<double>(y+21)).Store(y+21);
(s * sy220 + SIMD<double>(y+22)).Store(y+22);
(s * sy230 + SIMD<double>(y+23)).Store(y+23);
; }
template <TRIG_SIDE SIDE, TRIG_NORMAL NORM, ORDERING ORD, int DIM>
inline void KernelTriangularMult (size_t wx, double * pt, size_t dt, double * px, size_t dx);
template <TRIG_SIDE SIDE, TRIG_NORMAL NORM, ORDERING ORD, OPERATION OP, int DIM>
inline void KernelTriangularMultXY (size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
cerr << "missing implementation, side = " << SIDE << " norm = " << NORM
<< " ORD = " << ORD << ", OP = " << OP << ", DIM = " << DIM << endl; }
template <TRIG_SIDE SIDE, TRIG_NORMAL NORM, ORDERING ORD, int DIM>
inline void KernelTriangularSolve (size_t wx, double * pt, size_t dt, double * px, size_t dx);
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,ColMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularSolve<UpperRight,NonNormalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,ColMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularSolve<UpperRight,Normalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,ColMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularSolve<LowerLeft,NonNormalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,ColMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularSolve<LowerLeft,Normalized,RowMajor,0>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SET, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,ADD, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SUB, 0>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= U00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= U00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,ColMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= U00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= U00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,NonNormalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> Uinv00 = 1.0 / U00;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= Uinv00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= Uinv00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,ColMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,Normalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= L00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,ColMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= L00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,NonNormalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> Linv00 = 1.0 / L00;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
x0 *= Linv00;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
x0 *= Linv00;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,ColMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,Normalized,RowMajor,1>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SET, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,ADD, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SUB, 1>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 *= U00;
x0 += U01*x1;
x1 *= U11;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x1 *= U11;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,ColMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 *= U00;
x0 += U01*x1;
x1 *= U11;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x1 *= U11;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,NonNormalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> Uinv00 = 1.0 / U00;
SIMD<double> Uinv11 = 1.0 / U11;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 *= Uinv11;
x0 -= U01*x1;
x0 *= Uinv00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 *= Uinv11;
x0 -= U01*x1;
x0 *= Uinv00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 += U01*x1;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 += U01*x1;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,ColMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 += U01*x1;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 += U01*x1;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,Normalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 -= U01*x1;
x0.Store(px+0*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 -= U01*x1;
x0.Store(px+0*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U11, x1, y1);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,ColMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,NonNormalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> Linv00 = 1.0 / L00;
SIMD<double> Linv11 = 1.0 / L11;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,ColMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1+dt*0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,Normalized,RowMajor,2>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
x1 -= L10*x0;
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
x1 -= L10*x0;
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SET, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,ADD, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SUB, 2>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x1 *= U11;
x1 += U12*x2;
x2 *= U22;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x1 *= U11;
x1 += U12*x2;
x2 *= U22;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,ColMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x1 *= U11;
x1 += U12*x2;
x2 *= U22;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x1 *= U11;
x1 += U12*x2;
x2 *= U22;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,NonNormalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> Uinv00 = 1.0 / U00;
SIMD<double> Uinv11 = 1.0 / U11;
SIMD<double> Uinv22 = 1.0 / U22;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x2 *= Uinv22;
x1 -= U12*x2;
x1 *= Uinv11;
x0 -= U01*x1;
x0 -= U02*x2;
x0 *= Uinv00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x2 *= Uinv22;
x1 -= U12*x2;
x1 *= Uinv11;
x0 -= U01*x1;
x0 -= U02*x2;
x0 *= Uinv00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U12 = pt[1*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x0 += U01*x1;
x0 += U02*x2;
x1 += U12*x2;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x0 += U01*x1;
x0 += U02*x2;
x1 += U12*x2;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,ColMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U12 = pt[1+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x0 += U01*x1;
x0 += U02*x2;
x1 += U12*x2;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x0 += U01*x1;
x0 += U02*x2;
x1 += U12*x2;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,Normalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U12 = pt[1*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x1 -= U12*x2;
x0 -= U01*x1;
x0 -= U02*x2;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x1 -= U12*x2;
x0 -= U01*x1;
x0 -= U02*x2;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U22, x2, y2);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,ColMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,NonNormalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> Linv00 = 1.0 / L00;
SIMD<double> Linv11 = 1.0 / L11;
SIMD<double> Linv22 = 1.0 / L22;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x2 -= L20*x0;
x2 -= L21*x1;
x2 *= Linv22;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x2 -= L20*x0;
x2 -= L21*x1;
x2 *= Linv22;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,ColMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,Normalized,RowMajor,3>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
x1 -= L10*x0;
x2 -= L20*x0;
x2 -= L21*x1;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
x1 -= L10*x0;
x2 -= L20*x0;
x2 -= L21*x1;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SET, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,ADD, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SUB, 3>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 *= U11;
x1 += U12*x2;
x1 += U13*x3;
x2 *= U22;
x2 += U23*x3;
x3 *= U33;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 *= U11;
x1 += U12*x2;
x1 += U13*x3;
x2 *= U22;
x2 += U23*x3;
x3 *= U33;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,NonNormalized,ColMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U22 = pt[2+dt*2];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 *= U11;
x1 += U12*x2;
x1 += U13*x3;
x2 *= U22;
x2 += U23*x3;
x3 *= U33;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x0 *= U00;
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 *= U11;
x1 += U12*x2;
x1 += U13*x3;
x2 *= U22;
x2 += U23*x3;
x3 *= U33;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,NonNormalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U33 = pt[3*dt+3];
SIMD<double> Uinv00 = 1.0 / U00;
SIMD<double> Uinv11 = 1.0 / U11;
SIMD<double> Uinv22 = 1.0 / U22;
SIMD<double> Uinv33 = 1.0 / U33;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x3 *= Uinv33;
x2 -= U23*x3;
x2 *= Uinv22;
x1 -= U12*x2;
x1 -= U13*x3;
x1 *= Uinv11;
x0 -= U01*x1;
x0 -= U02*x2;
x0 -= U03*x3;
x0 *= Uinv00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x3 *= Uinv33;
x2 -= U23*x3;
x2 *= Uinv22;
x1 -= U12*x2;
x1 -= U13*x3;
x1 *= Uinv11;
x0 -= U01*x1;
x0 -= U02*x2;
x0 -= U03*x3;
x0 *= Uinv00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U22 = pt[2+dt*2];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U22 = pt[2+dt*2];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,RowMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0*dt+0];
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U11 = pt[1*dt+1];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U22 = pt[2*dt+2];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,NonNormalized,ColMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U00 = pt[0+dt*0];
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U11 = pt[1+dt*1];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U22 = pt[2+dt*2];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U23 = pt[2*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 += U12*x2;
x1 += U13*x3;
x2 += U23*x3;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 += U12*x2;
x1 += U13*x3;
x2 += U23*x3;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<UpperRight,Normalized,ColMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U23 = pt[2+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 += U12*x2;
x1 += U13*x3;
x2 += U23*x3;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x0 += U01*x1;
x0 += U02*x2;
x0 += U03*x3;
x1 += U12*x2;
x1 += U13*x3;
x2 += U23*x3;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<UpperRight,Normalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U23 = pt[2*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x2 -= U23*x3;
x1 -= U12*x2;
x1 -= U13*x3;
x0 -= U01*x1;
x0 -= U02*x2;
x0 -= U03*x3;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x2 -= U23*x3;
x1 -= U12*x2;
x1 -= U13*x3;
x0 -= U01*x1;
x0 -= U02*x2;
x0 -= U03*x3;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(U00, x0, y0);
FMAnonasm(U01, x1, y0);
FMAnonasm(U02, x2, y0);
FMAnonasm(U03, x3, y0);
FMAnonasm(U11, x1, y1);
FMAnonasm(U12, x2, y1);
FMAnonasm(U13, x3, y1);
FMAnonasm(U22, x2, y2);
FMAnonasm(U23, x3, y2);
FMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,RowMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0*dt+1];
SIMD<double> U02 = pt[0*dt+2];
SIMD<double> U03 = pt[0*dt+3];
SIMD<double> U12 = pt[1*dt+2];
SIMD<double> U13 = pt[1*dt+3];
SIMD<double> U23 = pt[2*dt+3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<UpperRight,Normalized,ColMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> U01 = pt[0+dt*1];
SIMD<double> U02 = pt[0+dt*2];
SIMD<double> U03 = pt[0+dt*3];
SIMD<double> U12 = pt[1+dt*2];
SIMD<double> U13 = pt[1+dt*3];
SIMD<double> U23 = pt[2+dt*3];
SIMD<double> U00(1.0);
SIMD<double> U11(1.0);
SIMD<double> U22(1.0);
SIMD<double> U33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(U00, x0, y0);
FNMAnonasm(U01, x1, y0);
FNMAnonasm(U02, x2, y0);
FNMAnonasm(U03, x3, y0);
FNMAnonasm(U11, x1, y1);
FNMAnonasm(U12, x2, y1);
FNMAnonasm(U13, x3, y1);
FNMAnonasm(U22, x2, y2);
FNMAnonasm(U23, x3, y2);
FNMAnonasm(U33, x3, y3);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x3 *= L33;
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x3 *= L33;
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,NonNormalized,ColMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x3 *= L33;
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x3 *= L33;
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 *= L22;
x2 += L20 * x0;
x2 += L21 * x1;
x1 *= L11;
x1 += L10 * x0;
x0 *= L00;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,NonNormalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L33 = pt[3*dt+3];
SIMD<double> Linv00 = 1.0 / L00;
SIMD<double> Linv11 = 1.0 / L11;
SIMD<double> Linv22 = 1.0 / L22;
SIMD<double> Linv33 = 1.0 / L33;
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x2 -= L20*x0;
x2 -= L21*x1;
x2 *= Linv22;
x3 -= L30*x0;
x3 -= L31*x1;
x3 -= L32*x2;
x3 *= Linv33;
x0.Store(px+0*dx+i); 
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x0 *= Linv00;
x1 -= L10*x0;
x1 *= Linv11;
x2 -= L20*x0;
x2 -= L21*x1;
x2 *= Linv22;
x3 -= L30*x0;
x3 -= L31*x1;
x3 -= L32*x2;
x3 *= Linv33;
x0.Store(px+0*dx+i, mask); 
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,RowMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0*dt+0];
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L11 = pt[1*dt+1];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L22 = pt[2*dt+2];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L33 = pt[3*dt+3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,NonNormalized,ColMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L00 = pt[0+dt*0];
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L11 = pt[1+dt*1];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L22 = pt[2+dt*2];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L33 = pt[3+dt*3];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMult<LowerLeft,Normalized,ColMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x3 += L30 * x0;
x3 += L31 * x1;
x3 += L32 * x2;
x2 += L20 * x0;
x2 += L21 * x1;
x1 += L10 * x0;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularSolve<LowerLeft,Normalized,RowMajor,4>(size_t wx, double * pt, size_t dt, double * px, size_t dx) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> x3(px+3*dx+i); 
x1 -= L10*x0;
x2 -= L20*x0;
x2 -= L21*x1;
x3 -= L30*x0;
x3 -= L31*x1;
x3 -= L32*x2;
x1.Store(px+1*dx+i); 
x2.Store(px+2*dx+i); 
x3.Store(px+3*dx+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
x1 -= L10*x0;
x2 -= L20*x0;
x2 -= L21*x1;
x3 -= L30*x0;
x3 -= L31*x1;
x3 -= L32*x2;
x1.Store(px+1*dx+i, mask); 
x2.Store(px+2*dx+i, mask); 
x3.Store(px+3*dx+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SET, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(0);
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(0);
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(0);
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(0);
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,ADD, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FMAnonasm(L30, x0, y3);
FMAnonasm(L31, x1, y3);
FMAnonasm(L32, x2, y3);
FMAnonasm(L33, x3, y3);
FMAnonasm(L20, x0, y2);
FMAnonasm(L21, x1, y2);
FMAnonasm(L22, x2, y2);
FMAnonasm(L10, x0, y1);
FMAnonasm(L11, x1, y1);
FMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,RowMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1*dt+0];
SIMD<double> L20 = pt[2*dt+0];
SIMD<double> L21 = pt[2*dt+1];
SIMD<double> L30 = pt[3*dt+0];
SIMD<double> L31 = pt[3*dt+1];
SIMD<double> L32 = pt[3*dt+2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
template <> 
inline void KernelTriangularMultXY<LowerLeft,Normalized,ColMajor,SUB, 4>(size_t wx, double * pt, size_t dt, double * px, size_t dx, double * py, size_t dy) { 
SIMD<double> L10 = pt[1+dt*0];
SIMD<double> L20 = pt[2+dt*0];
SIMD<double> L21 = pt[2+dt*1];
SIMD<double> L30 = pt[3+dt*0];
SIMD<double> L31 = pt[3+dt*1];
SIMD<double> L32 = pt[3+dt*2];
SIMD<double> L00(1.0);
SIMD<double> L11(1.0);
SIMD<double> L22(1.0);
SIMD<double> L33(1.0);
constexpr size_t SW = SIMD<double>::Size(); 
size_t i = 0; 
for ( ; i+SW <= wx; i+=SW) { 
SIMD<double> x0(px+0*dx+i); 
SIMD<double> y0(py+0*dy+i); 
SIMD<double> x1(px+1*dx+i); 
SIMD<double> y1(py+1*dy+i); 
SIMD<double> x2(px+2*dx+i); 
SIMD<double> y2(py+2*dy+i); 
SIMD<double> x3(px+3*dx+i); 
SIMD<double> y3(py+3*dy+i); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i); 
y1.Store(py+1*dy+i); 
y2.Store(py+2*dy+i); 
y3.Store(py+3*dy+i); 
}
size_t rest = wx % SW; 
if (rest == 0) return; 
SIMD<mask64> mask(rest);
SIMD<double> x0(px+0*dx+i, mask); 
SIMD<double> y0(py+0*dy+i, mask); 
SIMD<double> x1(px+1*dx+i, mask); 
SIMD<double> y1(py+1*dy+i, mask); 
SIMD<double> x2(px+2*dx+i, mask); 
SIMD<double> y2(py+2*dy+i, mask); 
SIMD<double> x3(px+3*dx+i, mask); 
SIMD<double> y3(py+3*dy+i, mask); 
FNMAnonasm(L30, x0, y3);
FNMAnonasm(L31, x1, y3);
FNMAnonasm(L32, x2, y3);
FNMAnonasm(L33, x3, y3);
FNMAnonasm(L20, x0, y2);
FNMAnonasm(L21, x1, y2);
FNMAnonasm(L22, x2, y2);
FNMAnonasm(L10, x0, y1);
FNMAnonasm(L11, x1, y1);
FNMAnonasm(L00, x0, y0);
y0.Store(py+0*dy+i, mask); 
y1.Store(py+1*dy+i, mask); 
y2.Store(py+2*dy+i, mask); 
y3.Store(py+3*dy+i, mask); 
}
